# sparks/models/model.py
from typing import (
    Dict,
    Tuple,
    Type,
    Generic,
    Protocol,
    Union,
    runtime_checkable,
    Any,
)
from abc import abstractmethod

from sparks.core import NDArrayType
from .state import StateType, SimulationData
from .reactions import RateType, RxnIDType, ReactionNetworkDef


@runtime_checkable
class BaseModelProtocol(Generic[RateType, StateType], Protocol):
    name: str

    RateClass: Type[RateType]
    StateClass: Type[StateType]


class ODEFunction(Protocol):
    """ODE Function for React models"""

    def __call__(
        self,
        dydt: StateType,
        k: RateType,
        t: NDArrayType,
        state: StateType,
        **kwargs: Any,
    ) -> StateType: ...


@runtime_checkable
class SupportsDeterministicODE(BaseModelProtocol, Protocol):

    @abstractmethod
    def ode_function(self) -> ODEFunction:
        """ODE function for deterministic simulation"""
        raise NotImplementedError


@runtime_checkable
class SupportsReactionNetwork(BaseModelProtocol, Protocol):
    RxnIDClass: Type[RxnIDType]

    @abstractmethod
    def reactions(self) -> ReactionNetworkDef:
        raise NotImplementedError


SupportsDeterministicSimulation = Union[
    SupportsDeterministicODE, SupportsReactionNetwork
]


@runtime_checkable
class SupportsStochasticSimulation(SupportsReactionNetwork, Protocol):

    # from React.simulators.stochastic import KMCConfig
    # from React.simulators.stochastic import SimulationResult
    # from sparks.simulators.stochastic
    class KMCConfig:
        pass

    from runkmc.simulation import SimulationResult

    @abstractmethod
    def get_kmc_inputs(
        self, k: RateType, s0: StateType, config: KMCConfig, model_name: str, **kwargs
    ) -> Dict[str, Any]:
        """Get the inputs for KMC simulation"""
        raise NotImplementedError

    @abstractmethod
    def parse_kmc_outputs(
        self, result: SimulationResult, model_name: str, **kwargs
    ) -> Tuple[SimulationData[StateType], Any]:
        """Parse the outputs from KMC simulation"""
        raise NotImplementedError



from __future__ import annotations
from typing import Dict, Type, Any

import numpy as np
from scipy.integrate import solve_ivp

from sparks.models import (
    RateType,
    StateType,
    SimulationData,
    ODEFunction,
    SupportsDeterministicODE,
    SupportsDeterministicSimulation,
    SupportsReactionNetwork,
)
from sparks.simulators import Simulator

from .compile import compile_odes, OdeSolutionState, OdeSolution
from .reactions import optimized_reaction_odes


def get_ode_function(
    model: SupportsDeterministicSimulation,
    prefer_reactions: bool = False,
) -> ODEFunction:

    has_custom_ode = isinstance(model, SupportsDeterministicODE)
    has_reactions = isinstance(model, SupportsReactionNetwork)

    if prefer_reactions and has_reactions:
        # Use reaction-generated ODE
        return optimized_reaction_odes(model.StateClass, model.reactions())
    elif has_custom_ode:
        # Use custom ODE function
        return model.ode_function()
    elif has_reactions:
        # Fallback to reaction-based ODE if no custom ODE is defined
        return optimized_reaction_odes(model.StateClass, model.reactions())
    else:
        raise ValueError(
            "Model must implement either SupportsDeterministicODE or SupportsReactionNetwork."
        )


def default_solver_params():
    return {
        "method": "BDF",
        "first_step": 1e-8,
        "atol": 1e-8,
        "rtol": 1e-6,
    }


class DeterministicSimulator(Simulator):
    """Simulator for deterministic systems of differential equations."""

    def __init__(
        self,
        model: SupportsDeterministicSimulation,
        k: RateType,
        prefer_reactions: bool = False,
    ):
        """
        Initialize the deterministic simulator with a model and rate constants.
        """
        super().__init__(model, k)
        self.model = model

        self.model_state_class: Type[StateType] = self.model.StateClass
        self.ode_fun = get_ode_function(self.model, prefer_reactions)

    def simulate(
        self,
        t_eval: np.ndarray,
        init_state: StateType,
        *,
        solver_kwargs: Dict[str, Any] = default_solver_params(),
        **model_kwargs: Any,
    ) -> SimulationData[StateType]:

        self.y0 = init_state
        y0_np = init_state.to_array()
        self.t_eval = t_eval

        ode_fun = compile_odes(
            self.model_state_class, self.ode_fun, self.k, **model_kwargs
        )

        self.sol = solve_ivp(
            fun=ode_fun,
            y0=y0_np,
            t_span=(t_eval[0], t_eval[-1]),
            t_eval=t_eval,
            dense_output=True,
            **solver_kwargs,
        )

        ode_sol: OdeSolution = self.sol.sol
        assert isinstance(
            ode_sol, OdeSolution
        ), "solve_ivp did not return a valid OdeSolution."
        self.ode_sol = OdeSolutionState(self.model_state_class, ode_sol)

        self.t = np.array(self.sol.t)
        self.y_np = np.array(self.sol.y)

        self.trajectory = self.model_state_class.from_array(self.y_np, t=self.t)

        self.data = SimulationData(trajectory=self.trajectory, init_state=init_state)

        return self.data



from functools import cached_property
from typing import Any, List, Tuple
from dataclasses import dataclass

import numpy as np

from sparks.core import NDArrayType, EPS
from sparks.models import (
    State,
    state_field,
    derived,
    SimulationData,
    ReactionNetworkDef,
    SupportsDeterministicODE,
    SupportsReactionNetwork,
    SupportsStochasticSimulation,
)
from sparks.simulators.stochastic import KMCConfig, SimulationResult


from .reactions import RxnID


@dataclass
class RateCoefficients:
    """Rate coefficients for binary uncontrolled free-radical copolymerization kinetics."""

    # Initiation
    kd: float = 0.0
    f: float = 1.0

    # Propagation
    kpAA: float = 0.0
    kpAB: float = 0.0
    kpBA: float = 0.0
    kpBB: float = 0.0

    # Depropagation
    kdAA: float = 0.0
    kdAB: float = 0.0
    kdBA: float = 0.0
    kdBB: float = 0.0

    # Termination
    ktcAA: float = 0.0
    ktcAB: float = 0.0
    ktcBB: float = 0.0

    ktdAA: float = 0.0
    ktdAB: float = 0.0
    ktdBB: float = 0.0

    def __post_init__(self):
        # Cache some commonly used values
        self._ktAA = self.ktcAA + self.ktdAA
        self._ktAB = self.ktcAB + self.ktdAB
        self._ktBB = self.ktcBB + self.ktdBB

    @cached_property
    def ktAA(self) -> float:
        """Total termination rate for AA"""
        return self._ktAA

    @cached_property
    def ktAB(self) -> float:
        """Total termination rate for AB"""
        return self._ktAB

    @cached_property
    def ktBB(self) -> float:
        """Total termination rate for BB"""
        return self._ktBB

    @cached_property
    def rA(self) -> float:
        """Reactivity ratio for monomer A"""
        return self.kpAA / self.kpAB if self.kpAB != 0 else 0

    @cached_property
    def rB(self) -> float:
        """Reactivity ratio for monomer B"""
        return self.kpBB / self.kpBA if self.kpBA != 0 else 0

    @cached_property
    def rX(self) -> float:
        """Cross-propagation ratio"""
        return self.kpAA / self.kpBB if self.kpBB != 0 else 0

    @cached_property
    def KAA(self) -> float:
        """Equilibrium constant for AA"""
        return self.kdAA / self.kpAA if self.kpAA != 0 else 0

    @cached_property
    def KAB(self) -> float:
        """Equilibrium constant for AB"""
        return self.kdAB / self.kpAB if self.kpAB != 0 else 0

    @cached_property
    def KBA(self) -> float:
        """Equilibrium constant for BA"""
        return self.kdBA / self.kpBA if self.kpBA != 0 else 0

    @cached_property
    def KBB(self) -> float:
        """Equilibrium constant for BB"""
        return self.kdBB / self.kpBB if self.kpBB != 0 else 0


@dataclass
class ModelState(State):
    """State for binary uncontrolled free-radical copolymerization kinetics."""

    I: NDArrayType = state_field(0.0)
    R: NDArrayType = state_field(0.0)
    A: NDArrayType = state_field(0.0)
    B: NDArrayType = state_field(0.0)
    RA: NDArrayType = state_field(0.0)
    RB: NDArrayType = state_field(0.0)
    PAA: NDArrayType = state_field(0.0)
    PAB: NDArrayType = state_field(0.0)
    PBA: NDArrayType = state_field(0.0)
    PBB: NDArrayType = state_field(0.0)
    PD: NDArrayType = state_field(0.0)

    def chain_species_ids(self) -> List[str]:
        return ["R", "RA", "RB", "PAA", "PAB", "PBA", "PBB", "PD"]

    @derived
    def M(self) -> NDArrayType:
        """Total monomer concentration"""
        return self.A + self.B

    @derived
    def fA(self) -> NDArrayType:
        """Fraction of monomer A"""
        return self.A / (self.M + EPS)

    @derived
    def fB(self) -> NDArrayType:
        """Fraction of monomer B"""
        return self.B / (self.M + EPS)

    @derived
    def PA(self) -> NDArrayType:
        """Concentration of polymer chains ending in ~~A"""
        return self.PAA + self.PBA

    @derived
    def PB(self) -> NDArrayType:
        """Concentration of polymer chains ending in ~~B"""
        return self.PAB + self.PBB

    @derived
    def P(self) -> NDArrayType:
        """Total polymer concentration"""
        return self.PA + self.PB

    @derived
    def pA(self) -> NDArrayType:
        """Fraction of polymer chains ending in ~~A"""
        return self.PA / (self.P + EPS)

    @derived
    def pB(self) -> NDArrayType:
        """Fraction of polymer chains ending in ~~B"""
        return self.PB / (self.P + EPS)

    @derived
    def pAA(self) -> NDArrayType:
        """Fraction of polymer chains ending in ~~AA"""
        return self.PAA / (self.PA + EPS)

    @derived
    def pAB(self) -> NDArrayType:
        """Fraction of polymer chains ending in ~~AB"""
        return self.PAB / (self.PB + EPS)

    @derived
    def pBA(self) -> NDArrayType:
        """Fraction of polymer chains ending in ~~BA"""
        return self.PBA / (self.PA + EPS)

    @derived
    def pBB(self) -> NDArrayType:
        """Fraction of polymer chains ending in ~~BB"""
        return self.PBB / (self.PB + EPS)

    def xA(self, s0: "ModelState") -> NDArrayType:
        """Conversion of monomer A"""
        return 1 - self.A / (s0.A + EPS)

    def xB(self, s0: "ModelState") -> NDArrayType:
        """Conversion of monomer B"""
        return 1 - self.B / (s0.B + EPS)

    def x(self, s0: "ModelState") -> NDArrayType:
        """Conversion of all monomer"""
        return 1 - (self.A + self.B) / (s0.A + s0.B + EPS)

    def FA(self, s0: "ModelState") -> NDArrayType:
        """Fraction of A monomers in polymer chains (composition)"""
        return (s0.A - self.A) / (s0.M - self.M + EPS)

    def FB(self, s0: "ModelState") -> NDArrayType:
        """Fraction of B monomers in polymer chains (composition)"""
        return (s0.B - self.B) / (s0.M - self.M + EPS)


@dataclass
class ChainModelState(State):
    """Represents the state variables for the chain length distribution model."""

    lam_0: NDArrayType = state_field(0.0)
    lam_1: NDArrayType = state_field(0.0)
    lam_2: NDArrayType = state_field(0.0)
    mu_0: NDArrayType = state_field(0.0)
    mu_1: NDArrayType = state_field(0.0)
    mu_2: NDArrayType = state_field(0.0)

    @derived
    def NACL(self) -> NDArrayType:
        """Number average chain length"""
        return (self.lam_1 + self.mu_1) / (self.lam_0 + self.mu_0 + EPS)

    @derived
    def WACL(self) -> NDArrayType:
        """Weight average chain length"""
        return (self.lam_2 + self.mu_2) / (self.lam_1 + self.mu_1 + EPS)

    @derived
    def PDI(self) -> NDArrayType:
        """Dispersity"""
        return self.WACL / self.NACL

    # @cached_property
    # def aP_1(self) -> StateVariable:
    #     """1st moment of the active CLD (excluding dead chains)"""
    #     return StateVariable("aP_1", (self.PA1 + self.PB1).value)

    # @cached_property
    # def P_1(self) -> StateVariable:
    #     """1st moment of the total CLD (active and dead chains)"""
    #     return StateVariable("P_1", (self.aP_1 + self.PD1).value)

    # @cached_property
    # def aP_2(self) -> StateVariable:
    #     """2nd moment of the active CLD (excluding dead chains)"""
    #     return StateVariable("aP_2", (self.PA2 + self.PB2).value)

    # @cached_property
    # def P_2(self) -> StateVariable:
    #     """2nd moment of the total CLD (active and dead chains)"""
    #     return StateVariable("P_2", (self.aP_2 + self.PD2).value)

    # @cached_property
    # def NACL(self) -> StateVariable:
    #     """Number average chain length"""
    #     return StateVariable("NACL", (self.P_1 / self.P_0).value)

    # @cached_property
    # def WACL(self) -> StateVariable:
    #     """Weight average chain length"""
    #     return StateVariable("WACL", (self.P_2 / self.P_1).value)

    # @cached_property
    # def PDI(self) -> StateVariable:
    #     """Dispersity"""
    #     return StateVariable("PDI", (self.WACL / self.NACL).value)


@dataclass
class SequenceModelState(State):
    """Represents the state variables for the sequence distribution model."""

    aSA0: NDArrayType = state_field()
    aSB0: NDArrayType = state_field()
    aSA1: NDArrayType = state_field()
    aSB1: NDArrayType = state_field()
    aSA2: NDArrayType = state_field()
    aSB2: NDArrayType = state_field()

    iSA0: NDArrayType = state_field()
    iSB0: NDArrayType = state_field()
    iSA1: NDArrayType = state_field()
    iSB1: NDArrayType = state_field()
    iSA2: NDArrayType = state_field()
    iSB2: NDArrayType = state_field()

    ###################################################
    ### Moments of the sequence length distribution ###
    ###################################################

    @derived
    def SA0(self) -> NDArrayType:
        """0th moment of the sequence length distribution"""
        return self.aSA0 + self.iSA0

    @derived
    def SB0(self) -> NDArrayType:
        """0th moment of the sequence length distribution"""
        return self.aSB0 + self.iSB0

    @derived
    def SA1(self) -> NDArrayType:
        """1st moment of the sequence length distribution"""
        return self.aSA1 + self.iSA1

    @derived
    def SB1(self) -> NDArrayType:
        """1st moment of the sequence length distribution"""
        return self.aSB1 + self.iSB1

    @derived
    def SA2(self) -> NDArrayType:
        """2nd moment of the sequence length distribution"""
        return self.aSA2 + self.iSA2

    @derived
    def SB2(self) -> NDArrayType:
        """2nd moment of the sequence length distribution"""
        return self.aSB2 + self.iSB2

    ##################################
    ### Active Sequence Statistics ###
    ##################################

    @derived
    def active_NASL_A(self) -> NDArrayType:
        """Number average sequence length of active A sequences"""
        return np.where(self.aSA0 > 0, self.aSA1 / self.aSA0, 0.0)

    @derived
    def active_NASL_B(self) -> NDArrayType:
        """Number average sequence length of active B sequences"""
        return np.where(self.aSB0 > 0, self.aSB1 / self.aSB0, 0.0)

    @derived
    def active_WASL_A(self) -> NDArrayType:
        """Weight average sequence length of active A sequences"""
        return np.where(self.aSA1 > 0, self.aSA2 / self.aSA1, 0.0)

    @derived
    def active_WASL_B(self) -> NDArrayType:
        """Weight average sequence length of active B sequences"""
        return np.where(self.aSB1 > 0, self.aSB2 / self.aSB1, 0.0)

    @derived
    def active_SDI_A(self) -> NDArrayType:
        """Sequence dispersity of active A sequences"""
        return np.where(
            self.active_NASL_A > 0, self.active_WASL_A / self.active_NASL_A, 0.0
        )

    @derived
    def active_SDI_B(self) -> NDArrayType:
        """Sequence dispersity of active B sequences"""
        return np.where(
            self.active_NASL_B > 0, self.active_WASL_B / self.active_NASL_B, 0.0
        )

    ####################################
    ### Inactive Sequence Statistics ###
    ####################################

    @derived
    def inactive_NASL_A(self) -> NDArrayType:
        """Number average sequence length of inactive A sequences"""
        return np.where(self.iSA0 > 0, self.iSA1 / self.iSA0, 0.0)

    @derived
    def inactive_NASL_B(self) -> NDArrayType:
        """Number average sequence length of inactive B sequences"""
        return np.where(self.iSB0 > 0, self.iSB1 / self.iSB0, 0.0)

    @derived
    def inactive_WASL_A(self) -> NDArrayType:
        """Weight average sequence length of inactive A sequences"""
        return np.where(self.iSA1 > 0, self.iSA2 / self.iSA1, 0.0)

    @derived
    def inactive_WASL_B(self) -> NDArrayType:
        """Weight average sequence length of inactive B sequences"""
        return np.where(self.iSB1 > 0, self.iSB2 / self.iSB1, 0.0)

    @derived
    def inactive_SDI_A(self) -> NDArrayType:
        """Sequence dispersity of inactive A sequences"""
        return np.where(
            self.inactive_NASL_A > 0, self.inactive_WASL_A / self.inactive_NASL_A, 0.0
        )

    @derived
    def inactive_SDI_B(self) -> NDArrayType:
        """Sequence dispersity of inactive B sequences"""
        return np.where(
            self.inactive_NASL_B > 0, self.inactive_WASL_B / self.inactive_NASL_B, 0.0
        )

    #################################
    ### Total Sequence Statistics ###
    #################################

    @derived
    def NASL_A(self) -> NDArrayType:
        """Number average sequence length of total A sequences (active + inactive)"""
        return np.where(self.SA0 > 0, self.SA1 / self.SA0, 0.0)

    @derived
    def NASL_B(self) -> NDArrayType:
        """Number average sequence length of total B sequences (active + inactive)"""
        return np.where(self.SB0 > 0, self.SB1 / self.SB0, 0.0)

    @derived
    def WASL_A(self) -> NDArrayType:
        """Weight average sequence length of total A sequences (active + inactive)"""
        return np.where(self.SA1 > 0, self.SA2 / self.SA1, 0.0)

    @derived
    def WASL_B(self) -> NDArrayType:
        """Weight average sequence length of total B sequences (active + inactive)"""
        return np.where(self.SB1 > 0, self.SB2 / self.SB1, 0.0)

    @derived
    def SDI_A(self) -> NDArrayType:
        """Sequence dispersity of total A sequences (active + inactive)"""
        return np.where(self.WASL_A > 0, self.NASL_A / self.WASL_A, 0.0)

    @derived
    def SDI_B(self) -> NDArrayType:
        """Sequence dispersity of total B sequences (active + inactive)"""
        return np.where(self.WASL_B > 0, self.NASL_B / self.WASL_B, 0.0)


class Model(
    SupportsDeterministicODE,
    SupportsStochasticSimulation,
    SupportsReactionNetwork,
):

    name = "FRP2"
    RateClass = RateCoefficients
    StateClass = ModelState
    RxnIDClass = RxnID

    def ode_function(self):
        from .equations import kinetic_odes

        return kinetic_odes

    def reactions(self) -> ReactionNetworkDef:
        from .reactions import reactions

        return reactions

    def get_kmc_inputs(
        self,
        k: RateCoefficients,
        state: ModelState,
        config: KMCConfig,
        model_name: str,
        **kwargs,
    ) -> dict[str, Any]:
        """Get KMC simulation inputs"""
        from .kmc import get_kmc_inputs

        return get_kmc_inputs(k, state, config, model_name)

    def parse_kmc_outputs(
        self, result: SimulationResult, model_name: str, **kwargs
    ) -> Tuple[SimulationData[ModelState], Any]:
        """Parse KMC simulation results"""
        from .kmc import parse_kmc_outputs

        return parse_kmc_outputs(result, model_name)


class ChainModel(SupportsDeterministicODE):

    name = "FRP2 Chain Model"
    RateClass = RateCoefficients
    StateClass = ChainModelState

    def ode_function(self):
        from .equations import chain_odes

        return chain_odes


class SequenceModel(SupportsDeterministicODE):

    name = "FRP2 Sequence Model"
    RateClass = RateCoefficients
    StateClass = SequenceModelState

    def ode_function(self):
        from .equations import sequence_odes

        return sequence_odes



from typing import Optional, Tuple, Dict

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from sparks.core import okprint, warnprint, dbgprint
from sparks.models import SimulationData
from sparks.simulators import DeterministicSimulator
from sparks.data.initiators import dTBPO
from sparks.data.monomers import alpha_methyl_styrene, methyl_methacrylate
from sparks.utils.formulas import C_to_K

from ..model import (
    Model,
    ModelState,
    RateCoefficients,
    SequenceModel,
    SequenceModelState,
)

##################################################
### HIGH TEMPERATURE AMS-MMA (115 C and 140 C) ###
##################################################

COLOR_115 = "#009ADE"
COLOR_140 = "#FF1F5B"


# ---- Experimental and model validation data ----

CONDS = {
    "T_115C": {"T_C": 115, "wAMS": 0.45, "wDTBPO": 0.02, "t_end_min": 1450},
    "T_140C": {"T_C": 140, "wAMS": 0.45, "wDTBPO": 0.02, "t_end_min": 1080},
}

# ---- Example Conditions ----


def k_AMS_MMA(temp_C: float) -> RateCoefficients:
    """Return the rate coefficients for the AMS MMA model at a given temperature in Celsius."""

    temp_K = C_to_K(temp_C)

    iprops = dTBPO(temp_K)
    mpropsA = alpha_methyl_styrene(temp_K)
    mpropsB = methyl_methacrylate(temp_K)

    kpAA = mpropsA.kp
    kpBB = mpropsB.kp

    dbgprint(f"Rate coefficients at {temp_C} C:")
    dbgprint(f"kp AMS: {kpAA/1e4:.4f}, kp MMA: {kpBB/1e4:.4f}")
    dbgprint(f"kt AMS: {mpropsA.kt/1e10:.4f}, kt MMA: {mpropsB.kt/1e10:.4f}")
    print()

    rA = 0.16
    rB = 0.41
    kpAB = kpAA / rA
    kpBA = kpBB / rB

    if temp_C == 115:
        KAA = 35
        KBB = 0.22
        qA = 8.2
        qB = 0.0
        kdBA = qA * kpAB
        kdAB = qB * kpBA
        KAB = kdAB / kpAB
        KBA = kdBA / kpBA
    elif temp_C == 140:
        KAA = 54
        KBB = 0.45
        qA = 19.1
        qB = 1.27
        kdBA = qA * kpAB
        kdAB = qB * kpBA
        KAB = kdAB / kpAB
        KBA = kdBA / kpBA
    else:
        raise ValueError(
            f"Unsupported temperature: {temp_C} C. Supported temperatures are 115 C and 140 C."
        )
    dbgprint(f"rA: {rA:.4f}, rB: {rB:.4f}")
    dbgprint(f"KAA: {KAA:.4f}, KAB: {KAB:.4f}, KBA: {KBA:.4f}, KBB: {KBB:.4f}")
    print()

    return RateCoefficients(
        T_C=temp_C,
        iprops=iprops,
        mpropsA=mpropsA,
        mpropsB=mpropsB,
        rA=rA,
        rB=rB,
        KAA=KAA,
        KAB=KAB,
        KBA=KBA,
        KBB=KBB,
    )


def c0_AMS_MMA(T_C: float, wAMS: float, wDTBPO: float) -> ModelState:
    """Create initial state for AMS-MMA simulation with specified weights."""
    temp_K = C_to_K(T_C)

    iprops = dTBPO(temp_K)
    mpropsA = alpha_methyl_styrene(temp_K)
    mpropsB = methyl_methacrylate(temp_K)

    wMMA = 1 - wAMS

    m_basis = 1000  # Basis in grams
    m_AMS = wAMS * (1 - wDTBPO) * m_basis
    m_MMA = wMMA * (1 - wDTBPO) * m_basis
    m_DTBP = wDTBPO * m_basis

    # Calculate moles
    mol_AMS = m_AMS / mpropsA.MW
    mol_MMA = m_MMA / mpropsB.MW
    mol_DTBP = m_DTBP / iprops.MW

    # Calculate volume in L
    v_AMS = m_AMS / mpropsA.dens_m
    v_MMA = m_MMA / mpropsB.dens_m
    v_total = v_AMS + v_MMA

    V_basis = v_total
    I = mol_DTBP
    A = mol_AMS
    B = mol_MMA

    return ModelState(V=V_basis, I=I, A=A, B=B)


# ---- Generate data for validation ----

solver_kwargs = dict(method="BDF", rtol=1e-8, atol=1e-8)


def get_sim_data() -> Tuple[DeterministicSimulator, DeterministicSimulator]:

    T_C = 115
    ds_115 = DeterministicSimulator(Model(), k=k_AMS_MMA(T_C))
    data_115 = ds_115.simulate(
        t_eval=np.linspace(0, 1450, 1000),
        init_state=c0_AMS_MMA(T_C, wAMS=0.45, wDTBPO=0.02),
        solver_kwargs=solver_kwargs,
    )

    T_C = 140
    ds_140 = DeterministicSimulator(Model(), k=k_AMS_MMA(T_C))
    data_140 = ds_140.simulate(
        t_eval=np.linspace(0, 1080, 1000),
        init_state=c0_AMS_MMA(T_C, wAMS=0.45, wDTBPO=0.02),
        solver_kwargs=solver_kwargs,
    )

    return ds_115, ds_140


def get_sim_sequence_data(
    ds_115: DeterministicSimulator, ds_140: DeterministicSimulator
) -> Tuple[DeterministicSimulator, DeterministicSimulator]:

    T_C = 115
    ds_115s = DeterministicSimulator(SequenceModel(), k=k_AMS_MMA(T_C))
    data_115 = ds_115s.simulate(
        t_eval=np.linspace(0, 1450, 1000),
        init_state=SequenceModelState(),
        solver_kwargs=solver_kwargs,
        s=ds_115.ode_sol,
    )

    T_C = 140
    ds_140s = DeterministicSimulator(SequenceModel(), k=k_AMS_MMA(T_C))
    data_140 = ds_140s.simulate(
        t_eval=np.linspace(0, 1080, 1000),
        init_state=SequenceModelState(),
        solver_kwargs=solver_kwargs,
        s=ds_140.ode_sol,
    )

    return ds_115s, ds_140s


# ---- Manuscript figures ----


def plot_conv_validation(
    data: Dict[str, SimulationData[ModelState]],
    exp_data: Optional[Dict[str, pd.DataFrame]] = None,
):
    """Plot conversion validation for all conditions."""

    fig, ax = plt.subplots(1, figsize=(3.5, 3.5), dpi=300)

    for i, (cond_name, sim_data) in enumerate(data.items()):

        condition = CONDS[cond_name]
        color = COLOR_115 if condition["T_C"] == 115 else COLOR_140

        cond_id = f"{int(100 * condition['wAMS'])} wt% AMS, {condition['wDTBPO'] * 100} wt% DTBPO"
        ax.plot(
            sim_data.trajectory.t,
            sim_data.trajectory.x(sim_data.init_state),
            label=cond_id,
            color=color,
        )

        if exp_data is None or cond_name not in exp_data:
            continue

        _exp_data = exp_data.get(cond_name)

        ax.plot(
            _exp_data["Time (min)"],
            _exp_data["Conversion (%)"],
            "^",
            color=color,
            alpha=0.5,
        )

    ax.tick_params(axis="both", direction="in", top=True, right=True)
    ax.set_xlim(0, 1600)
    ax.set_ylim(0, 0.8)
    ax.set_xticks([0, 400, 800, 1200, 1600])
    ax.set_xlabel("Time (min)")
    ax.set_ylabel("Total conversion")

    add_exp_ode_legend(ax, frameon=False, loc="lower right")

    return fig, ax


def plot_comp_validation(
    data: Dict[str, SimulationData[ModelState]],
    exp_data: Optional[Dict[str, pd.DataFrame]] = None,
):
    """Plot composition validation for all conditions."""
    fig, ax = plt.subplots(1, figsize=(3.5, 3.5), dpi=300)

    for i, (cond_name, sim_data) in enumerate(data.items()):

        condition = CONDS[cond_name]
        color = COLOR_115 if condition["T_C"] == 115 else COLOR_140

        cond_id = f"{int(100 * condition['wAMS'])} wt% AMS, {condition['wDTBPO'] * 100} wt% DTBPO"
        ax.plot(
            sim_data.trajectory.x(sim_data.init_state)[1:],
            sim_data.trajectory.FbarB[1:],
            label=cond_id,
            color=color,
        )

        if exp_data is None or cond_name not in exp_data:
            continue

        _exp_data = exp_data.get(cond_name)

        ax.plot(
            _exp_data["Conversion (%)"],
            1 - _exp_data["Composition (%)"],
            "^",
            color=color,
            alpha=0.5,
        )

    ax.tick_params(axis="both", direction="in", top=True, right=True)
    ax.set_ylim(0.5, 1.0)
    ax.set_xlim(0, 0.8)
    ax.set_xticks([0, 0.2, 0.4, 0.6, 0.8])
    ax.set_xticklabels(["0", "0.2", "0.4", "0.6", "0.8"])
    ax.set_xlabel("Total conversion")
    ax.set_ylabel("Copolymer Composition (MMA)")

    add_exp_ode_legend(ax, frameon=False, loc="lower left")

    return fig, ax


def plot_sequence_data(
    data: Dict[
        str, Tuple[SimulationData[ModelState], SimulationData[SequenceModelState]]
    ],
):
    """Plot composition validation for all conditions."""
    fig, ax = plt.subplots(1, figsize=(3.5, 3.5), dpi=300)

    for i, (cond_name, (sim_data, seq_data)) in enumerate(data.items()):

        condition = CONDS[cond_name]
        color = COLOR_115 if condition["T_C"] == 115 else COLOR_140

        cond_id = f"{int(100 * condition['wAMS'])} wt% AMS, {condition['wDTBPO'] * 100} wt% DTBPO"

        x_vals = sim_data.trajectory.x(sim_data.init_state)
        # end_idx = np.argmax(x_vals >= 0.59) if condition["T_C"] == 140 else None
        end_idx = None
        ax.plot(
            x_vals[1:end_idx],
            seq_data.trajectory.active_NASL_B[1:end_idx],
            label=cond_id,
            color=color,
        )

    ax.plot([0, 0.8], [1, 1], "k--")
    ax.tick_params(axis="both", direction="in", top=True, right=True)
    ax.set_ylim(0.5, 14.5)
    ax.set_xlim(0, 0.8)
    ax.set_xticks([0, 0.2, 0.4, 0.6, 0.8])
    ax.set_xticklabels(["0", "0.2", "0.4", "0.6", "0.8"])
    ax.set_yticks([2, 4, 6, 8, 10, 12, 14])
    ax.set_xlabel("Total conversion")
    ax.set_ylabel("Inst. Avg. Sequence Length (MMA)")

    return fig, ax


def add_exp_ode_legend(axs: Axes, **kwargs):

    markersize = kwargs.pop("markersize", 5)
    exp_dot = axs.plot(
        [], [], "k", label="Experiments", alpha=0.5, markersize=markersize, marker="^"
    )
    ode_line = axs.plot([], [], "k-", label="ODE", alpha=1.0)

    axs.legend(
        handles=[exp_dot[0], ode_line[0]],
        **kwargs,
    )

    return axs



from dataclasses import dataclass

import numpy as np

from sparks.core import NDArrayType
from sparks.core import NA, okprint, errprint
from sparks.data import MonomerType
from .model import RateCoefficients, ModelState


@dataclass
class Context:

    # Parameters
    use_diffusion_control: bool = True
    # use_diffusion_control: bool = False
    const_volume: bool = False
    const_temperature: bool = True

    # Initial state
    init_state: ModelState | None = None

    # Gel Effect
    triggered: bool = False
    t_event: NDArrayType | None = None
    Vf_cr1: float | None = None
    kt_cr: float | None = None
    Mw_cr: float | None = None

    debug_counter: int = 0
    debug_interval: int = 100

    # times = [10, 20, 30, 40, 50, 60, 70, 80, 90]
    times = [50, 100, 150, 200, 250, 300, 350, 400, 450, 500, 550, 600]
    time_idx: int = 0


def kinetic_odes(
    k: RateCoefficients, t: NDArrayType, s: ModelState, context: Context = Context()
) -> ModelState:
    """Defines the ODEs for the FRP1 model."""

    T_C = k.T_C
    T_K = T_C + 273.15  # Convert Celsius to Kelvin

    # Extract properties
    iprops = k.iprops
    mprops = k.mprops

    dydt = ModelState()
    dydt.t = t

    # Save initial state
    if not context.init_state and t >= 0:
        context.init_state = s

    nM0 = context.init_state.M  # Initial moles of monomer

    ########################
    ### Molecular Weight ###
    ########################

    MW_eff = mprops.MW
    Mn = MW_eff * s.mu_1 / s.mu_0
    Mw = MW_eff * s.mu_2 / s.mu_1

    mass_m = s.M * MW_eff
    mass_p = s.nPM * MW_eff
    mass_s = 0.0
    total_mass = mass_m + mass_p + mass_s
    wp = mass_p / total_mass

    vol_m = mass_m / mprops.dens_m
    vol_p = mass_p / mprops.dens_p
    vol_s = 0.0

    # Initiation
    kd = iprops.kd.get(mprops.id, iprops.kd[MonomerType.General])
    f = iprops.f.get(mprops.id, iprops.f[MonomerType.General])
    kth = mprops.kth

    # Propagation
    kp = mprops.kp

    # Chain transfer
    kfm = mprops.kf_m
    kfz = 0

    kfs = 0
    kfCTA = 0
    kfp = 0

    # Termination rate constants
    kt = mprops.kt
    ktd_ratio = mprops.ktd_ratio  # ktd / ktc

    kp_factor = 1
    kfm_factor = 1

    if context.use_diffusion_control:

        # Calculate the free volume
        Vf_m = (mprops.Vf_m + mprops.alpha_m * (T_K - mprops.Tg_m)) * vol_m / s.V
        Vf_p = (mprops.Vf_p + mprops.alpha_p * (T_K - mprops.Tg_p)) * vol_p / s.V
        Vf_s = 0.0
        Vf = Vf_m + Vf_p + Vf_s

        if not context.triggered:
            # 1) Segmental diffusion termination
            c = (nM0 - s.M) / s.V / MW_eff  # Mass concentration of accumulated polymer
            kt = kt * (1 + mprops.delta * c)

            K3_test = (Mw**mprops.m) * np.exp(mprops.A / Vf)
            if K3_test >= mprops.K3:
                context.triggered = True
                context.t_event = t
                context.Vf_cr1 = Vf * 1.0
                context.kt_cr = kt
                context.Mw_cr = Mw
                print(
                    f"K3 triggered at t={t:.2f} s: K3_test={K3_test:.2f}, K3={mprops.K3:.2f}"
                )
                print(f"Triggered at t={t:.2f} s, Vf={Vf:.2f} L, Mw={Mw/1e3:.2f} kDa")
        else:
            # print(f"t = {t:.2f} s, Vf = {Vf:.2f} L, Mw = {Mw/1e3:.2f} kDa")
            kt = (
                context.kt_cr
                * (context.Mw_cr / Mw) ** mprops.n
                * np.exp(-mprops.A * (1 / Vf - 1 / context.Vf_cr1))
            )

        # 2) Translational diffusion termination
        Vf_cr2 = mprops.Vf_c
        if Vf < Vf_cr2:
            kp_factor = kfm_factor = np.exp(-mprops.B * (1 / Vf - 1 / Vf_cr2))

        Vf_i = iprops.Vf_i.get(mprops.id, iprops.Vf_i[MonomerType.General])
        C_factor = iprops.C.get(mprops.id, iprops.C[MonomerType.General])
        if Vf < Vf_i:
            f = f * np.exp(-C_factor * (1 / Vf - 1 / Vf_i))

        # 3) Reaction diffusion termination
        D = (
            mprops.n_s * (mprops.l0) ** 2 * kp * s.M_conc / 6.0
        )  # Diffusion coefficient (m^2/s)
        delta = (6 * (vol_m / s.M) / (np.pi * NA)) ** (1 / 3.0)
        # delta = mprops.delta
        kt_rd = 8 * np.pi * NA * delta * D / 1000
        kt += 0.0 * kt_rd

    kp = kp * kp_factor
    kfm = kfm * kfm_factor

    ktc = kt / (1 + ktd_ratio)
    ktd = kt * ktd_ratio / (1 + ktd_ratio)

    kdp = 0
    kp = kp - kdp / s.M_conc if s.M_conc > 0 else kp

    Ri = 2 * f * kd * s.I_conc + 2 * kth * s.M_conc**3
    Rp = kp * s.M_conc * s.R_conc

    # Volume balance
    dydt.V = s.V * (-Rp * MW_eff * (1 / mprops.dens_m - 1 / mprops.dens_p))

    # Initiator balance
    dydt.I = s.V * (-kd * s.I_conc)

    # Monomer balance
    dydt.M = s.V * (
        # Consumption by propagation
        -kp * s.M_conc * s.R_conc
        # Consumption by chain transfer to monomer
        - kfm * s.R_conc * s.M_conc
        # Consumption by thermal initiation
        - 3 * kth * s.M_conc**3
    )

    dydt.nPM = s.V * Rp

    # Polymer radical balance
    dydt.R = s.V * (
        # Formation by initiation
        Ri
        # Consumption by chain transfer to impurity
        - kfz * s.R_conc * s.Z_conc
        # Consumption by termination
        - kt * s.R_conc**2
    )
    # Reactions for propagation and
    # chain transfer to monomer, solvent, and CTA
    # ommitted since they "consume" and "form" radicals

    # Dead polymer balance
    dydt.D = s.V * (
        # Formation by termination (comb.)
        +ktc * s.R_conc**2
        # Formation by termination (disp.)
        + 2 * ktd * s.R_conc**2
        # Formation by chain transfer to monomer
        + kfm * s.R_conc * s.M_conc
        # Formation by chain transfer to solvent
        + kfs * s.R_conc * s.S_conc
        # Formation by chain transfer to CTA
        + kfCTA * s.R_conc * s.CTA_conc
        # Formation by chain transfer to impurity
        + kfz * s.R_conc * s.Z_conc
    )

    # Solvent balance
    dydt.S = s.V * (
        # Consumption by chain transfer to solvent
        -kfs
        * s.R_conc
        * s.S_conc
    )

    # Chain transfer agent balance
    dydt.CTA = s.V * (
        # Consumption by chain transfer to CTA
        -kfCTA
        * s.R_conc
        * s.CTA_conc
    )

    # Impurity balance
    dydt.Z = s.V * (
        # Consumption by chain transfer to impurity
        -kfz
        * s.R_conc
        * s.Z_conc
    )

    # Active radicals (0th order moment)
    dydt.lam_0 = s.V * (
        # Formation by initiation
        Ri
        # Consumption by termination
        - kt * s.lam0_conc**2
        # Consumption from impurity
        - kfz * s.lam0_conc * s.Z_conc
    )

    # Active radicals (1st order moment)
    dydt.lam_1 = s.V * (
        # Formation by initiation
        Ri
        # Formation by propagation
        + kp * s.M_conc * s.lam0_conc
        # Consumption by termination
        - kt * s.lam0_conc * s.lam1_conc
        # Consumption from chain transfer
        - (kfCTA * s.CTA_conc + kfm * s.M_conc + kfs * s.S_conc)
        * (s.lam1_conc - s.lam0_conc)
        # Formation by chain transfer to polymer
        + kfp * (s.lam0_conc * s.mu2_conc - s.lam1_conc * s.mu1_conc)
        # Consumption by chain transfer to impurity
        - kfz * s.lam1_conc * s.Z_conc
    )

    # Active radicals (2nd order moment)
    dydt.lam_2 = s.V * (
        # Formation by initiation
        Ri
        # Formation by propagation
        + kp * s.M_conc * (s.lam0_conc + 2 * s.lam1_conc)
        # Consumption by termination
        - kt * s.lam0_conc * s.lam2_conc
        # Consumption from chain transfer
        - (kfCTA * s.CTA_conc + kfm * s.M_conc + kfs * s.S_conc)
        * (s.lam2_conc - s.lam0_conc)
        # Formation by chain transfer to polymer
        + kfp * (s.lam0_conc * s.mu3_conc - s.lam2_conc * s.mu1_conc)
        # Consumption by chain transfer to impurity
        - kfz * s.lam2_conc * s.Z_conc
    )

    # Dead radicals (0th order moment)
    dydt.mu_0 = s.V * (
        # Formation by termination
        +(0.5 * ktc + ktd) * s.lam0_conc**2
        # Formation by chain transfer
        + (kfm * s.M_conc + kfCTA * s.CTA_conc + kfs * s.S_conc + kfz * s.Z_conc)
        * s.lam0_conc
    )

    # Dead radicals (1st order moment)
    dydt.mu_1 = s.V * (
        # Formation by termination
        +(ktc + ktd) * s.lam0_conc * s.lam1_conc
        # Formation by chain transfer
        + (kfm * s.M_conc + kfCTA * s.CTA_conc + kfs * s.S_conc + kfz * s.Z_conc)
        * s.lam1_conc
        # Chain transfer to polymer
        + kfp * (s.lam1_conc * s.mu1_conc - s.lam0_conc * s.mu2_conc)
    )

    # Dead radicals (2nd order moment)
    dydt.mu_2 = s.V * (
        # Formation by termination (comb.)
        ktc * (s.lam0_conc * s.lam2_conc + s.lam1_conc**2)
        # Formation by termination (disp.)
        + ktd * s.lam0_conc * s.lam2_conc
        # Formation by chain transfer
        + (kfm * s.M_conc + kfCTA * s.CTA_conc + kfs * s.S_conc + kfz * s.Z_conc)
        * s.lam2_conc
        # Chain transfer to polymer
        + kfp * (s.lam2_conc * s.mu1_conc - s.lam0_conc * s.mu3_conc)
    )

    vol_error = 100 * abs((vol_m + vol_p - s.V) / s.V) if s.V > 0 else 0

    mole_error = 100 * abs((s.M + s.nPM - nM0) / (nM0)) if (nM0) > 0 else 0

    lam_0_error = (
        100 * abs((s.lam0_conc - s.R_conc) / s.lam0_conc) if s.lam0_conc > 0 else 0
    )

    # if context.debug_counter % context.debug_interval == 0:
    time = np.inf
    if context.time_idx < len(context.times):
        time = context.times[context.time_idx]

    if t >= time:
        context.time_idx += 1
        vol_error = 100 * abs((vol_m + vol_p + vol_s - s.V) / s.V)
        print(f"Step {context.time_idx}: t={t:.2f}")

        print(f"Time: {t:.2f} s, Volume: {s.V:.3e} L")
        print(
            f"kd: {kd:.3e}, kp: {kp:.3e}, kt: {kt:.3e}, kfm: {kfm:.3e}, kfs: {kfs:.3e}"
        )
        print(f"Vf: {Vf:.3e} L, Mn: {Mn:.3e} g/mol, Mw: {Mw:.3e} g/mol")

        mole_error_str = f"\tMole balance error: {mole_error:.3e}%"
        if mole_error > 1.0:
            errprint(mole_error_str)
        else:
            okprint(mole_error_str)

        vol_error_str = f"\tVolume balance error: {vol_error:.3e}%"
        print(f"\tVolume: {s.V:.3f} L, vol_m: {vol_m:.3f} L, vol_p: {vol_p:.3f} L")
        if vol_error > 1.0:
            errprint(vol_error_str)
        else:
            okprint(vol_error_str)

        lam_0_error_str = f"\tlam_0 error: {lam_0_error:.3e}%"
        if lam_0_error > 1.0:
            errprint(lam_0_error_str)
            errprint(f"\t\tkt: {kt:.3e}, ktd: {ktd:.3e}, ktc: {ktc:.3e}")
            errprint(
                f"\t\tRi: {Ri:.3e}, Rp: {Rp:.3e}, M: {s.M_conc:.3e}, R: {s.R_conc:.3e}"
            )
            errprint(f"\t\tCalculated lam_0: {s.lam_0:.3e}, Expected: {s.R_conc:.3e}")
        else:
            okprint(lam_0_error_str)
        # print(f"lam_0: {s.lam_0.value:.3e}, Ptot: {Ptot:.3e}, Rtot: {Rtot:.3e}")
        active_NACL = (s.lam_1 / s.lam_0) if s.lam_0 > 0 else 0
        print(f"\tActive NACL: {active_NACL:.2f}")

        inactive_NACL = (s.mu_1 / s.mu_0) if s.mu_0 > 0 else 0
        print(f"\tInactive NACL: {inactive_NACL:.2f}")

        print()
    context.debug_counter += 1

    return dydt



from dataclasses import dataclass

import numpy as np

from sparks.core import NDArrayType, EPS
from sparks.models import State, state_field, derived
from sparks.data import InitiatorData, MonomerData
from sparks.models import SupportsDeterministicODE, ODEFunction


@dataclass
class RateCoefficients:

    mpropsA: MonomerData
    mpropsB: MonomerData
    iprops: InitiatorData
    T_C: float = 0.0

    # Copolymer reactivity
    rA: float = 1.0
    rB: float = 1.0
    KAA: float = 0.0
    KAB: float = 0.0
    KBA: float = 0.0
    KBB: float = 0.0

    def kp(self, f: float = 1.0):
        """Calculate propagation rate coefficients"""
        kpAA = f * self.mpropsA.kp
        kpBB = f * self.mpropsB.kp
        kpAB = kpAA / self.rA
        kpBA = kpBB / self.rB
        return kpAA, kpAB, kpBA, kpBB

    def kd(self, f: float = 1.0):
        """Calculate depropagation rate coefficients"""
        kpAA, kpAB, kpBA, kpBB = self.kp(f=f)
        kdAA = self.KAA * kpAA
        kdAB = self.KAB * kpAB
        kdBA = self.KBA * kpBA
        kdBB = self.KBB * kpBB
        return kdAA, kdAB, kdBA, kdBB

    def kfm(self, f: float = 1.0):
        """Calculate transfer to monomer rate coefficients"""
        kf_mAA = f * self.mpropsA.kf_m
        kf_mBB = f * self.mpropsB.kf_m
        kf_mAB = kf_mAA / self.rB
        kf_mBA = kf_mBB / self.rA
        return kf_mAA, kf_mAB, kf_mBA, kf_mBB


@dataclass
class ModelState(State):

    V: NDArrayType = state_field(1.0)

    I: NDArrayType = state_field()
    R: NDArrayType = state_field()
    A: NDArrayType = state_field()
    B: NDArrayType = state_field()
    RA: NDArrayType = state_field()
    RB: NDArrayType = state_field()
    PAA: NDArrayType = state_field()
    PAB: NDArrayType = state_field()
    PBA: NDArrayType = state_field()
    PBB: NDArrayType = state_field()
    D: NDArrayType = state_field()

    # Moles of monomers converted to polymer
    nPA: NDArrayType = state_field()
    nPB: NDArrayType = state_field()

    CTA: NDArrayType = state_field()
    S: NDArrayType = state_field()
    Z: NDArrayType = state_field()

    # Live Radical Moment Equations
    lam_0: NDArrayType = state_field()
    lam_1: NDArrayType = state_field()
    lam_2: NDArrayType = state_field()

    # Dead Polymer Moment Equations
    mu_0: NDArrayType = state_field()
    mu_1: NDArrayType = state_field()
    mu_2: NDArrayType = state_field()

    ##############################
    ### Species concentrations ###
    ##############################

    @derived
    def I_conc(self) -> NDArrayType:
        """Concentration of initiator"""
        return np.where(self.V > 0, self.I / self.V, 0.0)

    @derived
    def R_conc(self) -> NDArrayType:
        """Concentration of free radicals"""
        return np.where(self.V > 0, self.R / self.V, 0.0)

    @derived
    def RA_conc(self) -> NDArrayType:
        """Concentration of R-A dyads"""
        return np.where(self.V > 0, self.RA / self.V, 0.0)

    @derived
    def RB_conc(self) -> NDArrayType:
        """Concentration of R-B dyads"""
        return np.where(self.V > 0, self.RB / self.V, 0.0)

    @derived
    def A_conc(self) -> NDArrayType:
        """Concentration of monomer A"""
        return np.where(self.V > 0, self.A / self.V, 0.0)

    @derived
    def B_conc(self) -> NDArrayType:
        """Concentration of monomer B"""
        return np.where(self.V > 0, self.B / self.V, 0.0)

    @derived
    def M_conc(self) -> NDArrayType:
        """Concentration of monomers"""
        return np.where(self.V > 0, self.M / self.V, 0.0)

    @derived
    def PAA_conc(self) -> NDArrayType:
        """Concentration of active polymer chains ending in AA"""
        return np.where(self.V > 0, self.PAA / self.V, 0.0)

    @derived
    def PAB_conc(self) -> NDArrayType:
        """Concentration of active polymer chains ending in AB"""
        return np.where(self.V > 0, self.PAB / self.V, 0.0)

    @derived
    def PBA_conc(self) -> NDArrayType:
        """Concentration of active polymer chains ending in BA"""
        return np.where(self.V > 0, self.PBA / self.V, 0.0)

    @derived
    def PBB_conc(self) -> NDArrayType:
        """Concentration of active polymer chains ending in BB"""
        return np.where(self.V > 0, self.PBB / self.V, 0.0)

    @derived
    def PA_conc(self) -> NDArrayType:
        """Concentration of active polymer chains ending in A"""
        return self.PAA_conc + self.PBA_conc

    @derived
    def PB_conc(self) -> NDArrayType:
        """Concentration of active polymer chains ending in B"""
        return self.PAB_conc + self.PBB_conc

    @derived
    def P_conc(self) -> NDArrayType:
        """Total concentration of active polymer chains"""
        return self.PA_conc + self.PB_conc

    @derived
    def S_conc(self) -> NDArrayType:
        """Concentration of solvent"""
        return np.where(self.V > 0, self.S / self.V, 0.0)

    @derived
    def Z_conc(self) -> NDArrayType:
        """Concentration of impurity"""
        return np.where(self.V > 0, self.Z / self.V, 0.0)

    @derived
    def CTA_conc(self) -> NDArrayType:
        """Concentration of chain transfer agent"""
        return np.where(self.V > 0, self.CTA / self.V, 0.0)

    @derived
    def lam_0_conc(self) -> NDArrayType:
        """Concentration of living radical moment 0"""
        return np.where(self.V > 0, self.lam_0 / self.V, 0.0)

    @derived
    def lam_1_conc(self) -> NDArrayType:
        """Concentration of living radical moment 1"""
        return np.where(self.V > 0, self.lam_1 / self.V, 0.0)

    @derived
    def lam_2_conc(self) -> NDArrayType:
        """Concentration of living radical moment 2"""
        return np.where(self.V > 0, self.lam_2 / self.V, 0.0)

    @derived
    def mu_0_conc(self) -> NDArrayType:
        """Concentration of dead polymer moment 0"""
        return np.where(self.V > 0, self.mu_0 / self.V, 0.0)

    @derived
    def mu_1_conc(self) -> NDArrayType:
        """Concentration of dead polymer moment 1"""
        return np.where(self.V > 0, self.mu_1 / self.V, 0.0)

    @derived
    def mu_2_conc(self) -> NDArrayType:
        """Concentration of dead polymer moment 2"""
        return np.where(self.V > 0, self.mu_2 / self.V, 0.0)

    ##########################
    ### Monomer Properties ###
    ##########################

    @derived
    def M(self) -> NDArrayType:
        """Total moles of monomer"""
        return self.A + self.B

    @derived
    def fA(self) -> NDArrayType:
        """Fraction of monomer A"""
        return np.where(self.M > 0, self.A / self.M, 0.0)

    @derived
    def fB(self) -> NDArrayType:
        """Fraction of monomer B"""
        return np.where(self.M > 0, self.B / self.M, 0.0)

    #################################
    ### Active Polymer Properties ###
    #################################

    @derived
    def PA(self) -> NDArrayType:
        """Moles of active polymer chains ending in A"""
        return self.PAA + self.PBA

    @derived
    def PB(self) -> NDArrayType:
        """Moles of active polymer chains ending in B"""
        return self.PAB + self.PBB

    @derived
    def P(self) -> NDArrayType:
        """Total moles of active polymer chains"""
        return self.PA + self.PB

    @derived
    def pA(self) -> NDArrayType:
        """Fraction of active polymer chains ending in A"""
        return self.PA / (self.P + EPS)

    @derived
    def pB(self) -> NDArrayType:
        """Fraction of active polymer chains ending in B"""
        return self.PB / (self.P + EPS)

    @derived
    def pAA(self) -> NDArrayType:
        """Fraction of active ~A chains ending in ~AA"""
        return self.PAA / (self.PA + EPS)

    @derived
    def pAB(self) -> NDArrayType:
        """Fraction of active ~B chains ending in ~AB"""
        return self.PAB / (self.PB + EPS)

    @derived
    def pBA(self) -> NDArrayType:
        """Fraction of active ~A chains ending in ~BA"""
        return self.PBA / (self.PA + EPS)

    @derived
    def pBB(self) -> NDArrayType:
        """Fraction of active ~B chains ending in ~BB"""
        return self.PBB / (self.PB + EPS)

    @derived
    def FA(self) -> NDArrayType:
        """Composition of monomer A in the active polymer"""
        return self.PA / (self.P + EPS)

    @derived
    def FB(self) -> NDArrayType:
        """Composition of monomer B in the active polymer"""
        return self.PB / (self.P + EPS)

    @derived
    def beta(self) -> NDArrayType:
        """Ratio of monomer B to A in the active polymer"""
        return self.PB / (self.PA + EPS)

    #####################################
    ### Cumulative Polymer Properties ###
    #####################################

    @derived
    def nP(self) -> NDArrayType:
        """Total moles of polymer"""
        return self.nPA + self.nPB

    @derived
    def FbarA(self) -> NDArrayType:
        """Cumulative composition of monomer A in the polymer"""
        return self.nPA / (self.nP + EPS)

    @derived
    def FbarB(self) -> NDArrayType:
        """Cumulative composition of monomer B in the polymer"""
        return self.nPB / (self.nP + EPS)

    # def MW_Eff(self, k: RateCoefficients) -> NDArrayType:
    #     """Molecular weight of the polymer"""
    #     return self.FbarA * k.mpropsA.MW + self.FbarB * k.mpropsB.MW

    ###############################
    ### Chain Length Properties ###
    ###############################

    @derived
    def cld_m0(self) -> NDArrayType:
        """0th moment of the chain length distribution"""
        return 0 * self.lam_0 + self.mu_0

    @derived
    def cld_m1(self) -> NDArrayType:
        """1st moment of the chain length distribution"""
        return 0 * self.lam_1 + self.mu_1

    @derived
    def cld_m2(self) -> NDArrayType:
        """2nd moment of the chain length distribution"""
        return 0 * self.lam_2 + self.mu_2

    @derived
    def active_NACL(self) -> NDArrayType:
        """Number average chain length"""
        return np.where(self.lam_0 > 0, self.lam_1 / self.lam_0, 0.0)

    @derived
    def active_WACL(self) -> NDArrayType:
        """Weight average chain length"""
        return np.where(self.lam_1 > 0, self.lam_2 / self.lam_1, 0.0)

    @derived
    def active_PDI(self) -> NDArrayType:
        """Polydispersity index"""
        return np.where(self.active_NACL > 0, self.active_WACL / self.active_NACL, 0.0)

    @derived
    def NACL(self) -> NDArrayType:
        """Number average chain length"""
        return np.where(self.cld_m0 > 0, self.cld_m1 / self.cld_m0, 0.0)

    @derived
    def WACL(self) -> NDArrayType:
        """Weight average chain length"""
        return np.where(self.cld_m1 > 0, self.cld_m2 / self.cld_m1, 0.0)

    @derived
    def PDI(self) -> NDArrayType:
        """Polydispersity index"""
        return np.where(self.NACL > 0, self.WACL / self.NACL, 0.0)

    ###########################
    ### Computed Properties ###
    ###########################

    def xA(self, s0: "ModelState") -> NDArrayType:
        """Conversion of monomer A"""
        return np.where(s0.A > 0, 1 - self.A / s0.A, 0.0)

    def xB(self, s0: "ModelState") -> NDArrayType:
        """Conversion of monomer B"""
        return np.where(s0.B > 0, 1 - self.B / s0.B, 0.0)

    def x(self, s0: "ModelState") -> NDArrayType:
        """Conversion of all monomer"""
        return np.where((s0.A + s0.B) > 0, 1 - (self.A + self.B) / (s0.A + s0.B), 0.0)


@dataclass
class SequenceModelState(State):
    """Represents the state variables for the sequence distribution model."""

    aSA0: NDArrayType = state_field()
    aSB0: NDArrayType = state_field()
    aSA1: NDArrayType = state_field()
    aSB1: NDArrayType = state_field()
    aSA2: NDArrayType = state_field()
    aSB2: NDArrayType = state_field()

    iSA0: NDArrayType = state_field()
    iSB0: NDArrayType = state_field()
    iSA1: NDArrayType = state_field()
    iSB1: NDArrayType = state_field()
    iSA2: NDArrayType = state_field()
    iSB2: NDArrayType = state_field()

    ###################################################
    ### Moments of the sequence length distribution ###
    ###################################################

    @derived
    def SA0(self) -> NDArrayType:
        """0th moment of the sequence length distribution"""
        return self.aSA0 + self.iSA0

    @derived
    def SB0(self) -> NDArrayType:
        """0th moment of the sequence length distribution"""
        return self.aSB0 + self.iSB0

    @derived
    def SA1(self) -> NDArrayType:
        """1st moment of the sequence length distribution"""
        return self.aSA1 + self.iSA1

    @derived
    def SB1(self) -> NDArrayType:
        """1st moment of the sequence length distribution"""
        return self.aSB1 + self.iSB1

    @derived
    def SA2(self) -> NDArrayType:
        """2nd moment of the sequence length distribution"""
        return self.aSA2 + self.iSA2

    @derived
    def SB2(self) -> NDArrayType:
        """2nd moment of the sequence length distribution"""
        return self.aSB2 + self.iSB2

    ##################################
    ### Active Sequence Statistics ###
    ##################################

    @derived
    def active_NASL_A(self) -> NDArrayType:
        """Number average sequence length of active A sequences"""
        return np.where(self.aSA0 > 0, self.aSA1 / self.aSA0, 0.0)

    @derived
    def active_NASL_B(self) -> NDArrayType:
        """Number average sequence length of active B sequences"""
        return np.where(self.aSB0 > 0, self.aSB1 / self.aSB0, 0.0)

    @derived
    def active_WASL_A(self) -> NDArrayType:
        """Weight average sequence length of active A sequences"""
        return np.where(self.aSA1 > 0, self.aSA2 / self.aSA1, 0.0)

    @derived
    def active_WASL_B(self) -> NDArrayType:
        """Weight average sequence length of active B sequences"""
        return np.where(self.aSB1 > 0, self.aSB2 / self.aSB1, 0.0)

    @derived
    def active_SDI_A(self) -> NDArrayType:
        """Sequence dispersity of active A sequences"""
        return np.where(
            self.active_NASL_A > 0, self.active_WASL_A / self.active_NASL_A, 0.0
        )

    @derived
    def active_SDI_B(self) -> NDArrayType:
        """Sequence dispersity of active B sequences"""
        return np.where(
            self.active_NASL_B > 0, self.active_WASL_B / self.active_NASL_B, 0.0
        )

    ####################################
    ### Inactive Sequence Statistics ###
    ####################################

    @derived
    def inactive_NASL_A(self) -> NDArrayType:
        """Number average sequence length of inactive A sequences"""
        return np.where(self.iSA0 > 0, self.iSA1 / self.iSA0, 0.0)

    @derived
    def inactive_NASL_B(self) -> NDArrayType:
        """Number average sequence length of inactive B sequences"""
        return np.where(self.iSB0 > 0, self.iSB1 / self.iSB0, 0.0)

    @derived
    def inactive_WASL_A(self) -> NDArrayType:
        """Weight average sequence length of inactive A sequences"""
        return np.where(self.iSA1 > 0, self.iSA2 / self.iSA1, 0.0)

    @derived
    def inactive_WASL_B(self) -> NDArrayType:
        """Weight average sequence length of inactive B sequences"""
        return np.where(self.iSB1 > 0, self.iSB2 / self.iSB1, 0.0)

    @derived
    def inactive_SDI_A(self) -> NDArrayType:
        """Sequence dispersity of inactive A sequences"""
        return np.where(
            self.inactive_NASL_A > 0, self.inactive_WASL_A / self.inactive_NASL_A, 0.0
        )

    @derived
    def inactive_SDI_B(self) -> NDArrayType:
        """Sequence dispersity of inactive B sequences"""
        return np.where(
            self.inactive_NASL_B > 0, self.inactive_WASL_B / self.inactive_NASL_B, 0.0
        )

    #################################
    ### Total Sequence Statistics ###
    #################################

    @derived
    def NASL_A(self) -> NDArrayType:
        """Number average sequence length of total A sequences (active + inactive)"""
        return np.where(self.SA0 > 0, self.SA1 / self.SA0, 0.0)

    @derived
    def NASL_B(self) -> NDArrayType:
        """Number average sequence length of total B sequences (active + inactive)"""
        return np.where(self.SB0 > 0, self.SB1 / self.SB0, 0.0)

    @derived
    def WASL_A(self) -> NDArrayType:
        """Weight average sequence length of total A sequences (active + inactive)"""
        return np.where(self.SA1 > 0, self.SA2 / self.SA1, 0.0)

    @derived
    def WASL_B(self) -> NDArrayType:
        """Weight average sequence length of total B sequences (active + inactive)"""
        return np.where(self.SB1 > 0, self.SB2 / self.SB1, 0.0)

    @derived
    def SDI_A(self) -> NDArrayType:
        """Sequence dispersity of total A sequences (active + inactive)"""
        return np.where(self.WASL_A > 0, self.NASL_A / self.WASL_A, 0.0)

    @derived
    def SDI_B(self) -> NDArrayType:
        """Sequence dispersity of total B sequences (active + inactive)"""
        return np.where(self.WASL_B > 0, self.NASL_B / self.WASL_B, 0.0)


def get_sms(s: ModelState, k: RateCoefficients) -> SequenceModelState:

    kpAA, kpAB, kpBA, kpBB = k.kp()
    kdAA, kdAB, kdBA, kdBB = k.kd()

    RpAA = kpAA * s.PA_conc * s.A_conc - kdAA * s.PAA_conc
    RpAB = kpAB * s.PA_conc * s.B_conc - kdAB * s.PAB_conc
    RpBA = kpBA * s.PB_conc * s.A_conc - kdBA * s.PBA_conc
    RpBB = kpBB * s.PB_conc * s.B_conc - kdBB * s.PBB_conc

    Rp = RpAA + RpAB + RpBA + RpBB
    RpA = RpAA + RpAB
    RpB = RpBA + RpBB

    paa = RpAA / RpA
    pbb = RpBB / RpB

    # print(f"RpAA: {RpAA}, RpAB: {RpAB}, RpA: {RpA}")
    # print(f"RpBA: {RpBA}, RpBB: {RpBB}, RpB: {RpB}")
    # print(f"paa: {paa}, pbb: {pbb}")

    sms = SequenceModelState(t=s.t)

    sms.aSA0 = np.ones_like(paa)
    sms.aSB0 = np.ones_like(pbb)
    sms.aSA1 = 1 / (1 - paa)
    sms.aSB1 = 1 / (1 - pbb)
    sms.aSA2 = (1 + paa) / (1 - paa) ** 2
    sms.aSB2 = (1 + pbb) / (1 - pbb) ** 2

    return sms


class Model(SupportsDeterministicODE):

    name = "FRP2 ODE"
    RateClass = RateCoefficients
    StateClass = ModelState

    _state = ModelState()

    def ode_function(self) -> ODEFunction:
        from .equations import kinetic_odes

        return kinetic_odes


class SequenceModel(SupportsDeterministicODE):

    name = "FRP2 ODE Sequence"
    RateClass = RateCoefficients
    StateClass = SequenceModelState

    _state = SequenceModelState()

    def ode_function(self) -> ODEFunction:
        from .equations import sequence_odes

        return sequence_odes



from .model import RateCoefficients, ModelState, Model
from .equations import kinetic_odes, Context

__all__ = [
    "RateCoefficients",
    "ModelState",
    "Model",
    "kinetic_odes",
    "Context",
]



from typing import List
from dataclasses import dataclass
from functools import cached_property

from sparks.core import NDArrayType, EPS
from sparks.models import State, state_field, SupportsDeterministicODE
from sparks.data import InitiatorData, MonomerData


@dataclass
class RateCoefficients:
    """Rate constants for homopolymerization kinetics."""

    iprops: InitiatorData  # Initiator properties
    mprops: MonomerData  # Monomer properties
    T_C: float = 0.0  # Temperature in Celsius


@dataclass
class ModelState(State):
    """Represents the state variables for the homopolymer model."""

    V: NDArrayType = state_field(1)

    I: NDArrayType = state_field()
    M: NDArrayType = state_field()
    R: NDArrayType = state_field()
    D: NDArrayType = state_field()

    # Number of moles of monomer in polymer
    nPM: NDArrayType = state_field()

    CTA: NDArrayType = state_field()
    S: NDArrayType = state_field()
    Z: NDArrayType = state_field()

    # Live Radical Moment Equations
    lam_0: NDArrayType = state_field()
    lam_1: NDArrayType = state_field()
    lam_2: NDArrayType = state_field()

    # Dead Polymer Moment Equations
    mu_0: NDArrayType = state_field()
    mu_1: NDArrayType = state_field()
    mu_2: NDArrayType = state_field()

    def chain_species_ids(self) -> List[str]:
        return ["R", "D"]

    @cached_property
    def I_conc(self) -> NDArrayType:
        """Initiator concentration"""
        return self.I / (self.V + EPS)

    @cached_property
    def M_conc(self) -> NDArrayType:
        """Monomer concentration"""
        return self.M / (self.V + EPS)

    @cached_property
    def R_conc(self) -> NDArrayType:
        """Radical concentration"""
        return self.R / (self.V + EPS)

    @cached_property
    def D_conc(self) -> NDArrayType:
        """Dead polymer concentration"""
        return self.D / (self.V + EPS)

    @cached_property
    def CTA_conc(self) -> NDArrayType:
        """Chain transfer agent concentration"""
        return self.CTA / (self.V + EPS)

    @cached_property
    def Z_conc(self) -> NDArrayType:
        """Impurity concentration"""
        return self.Z / (self.V + EPS)

    @cached_property
    def S_conc(self) -> NDArrayType:
        """Solvent concentration"""
        return self.S / (self.V + EPS)

    @cached_property
    def lam0_conc(self) -> NDArrayType:
        """Live radical moment 0 concentration"""
        return self.lam_0 / (self.V + EPS)

    @cached_property
    def lam1_conc(self) -> NDArrayType:
        """Live radical moment 1 concentration"""
        return self.lam_1 / (self.V + EPS)

    @cached_property
    def lam2_conc(self) -> NDArrayType:
        """Live radical moment 2 concentration"""
        return self.lam_2 / (self.V + EPS)

    @cached_property
    def mu0_conc(self) -> NDArrayType:
        """Dead polymer moment 0 concentration"""
        return self.mu_0 / (self.V + EPS)

    @cached_property
    def mu1_conc(self) -> NDArrayType:
        """Dead polymer moment 1 concentration"""
        return self.mu_1 / (self.V + EPS)

    @cached_property
    def mu2_conc(self) -> NDArrayType:
        """Dead polymer moment 2 concentration"""
        return self.mu_2 / (self.V + EPS)

    @cached_property
    def mu_3(self) -> NDArrayType:
        """Dead polymer moment 3"""
        return (
            (
                self.mu_2
                / (self.mu_0 * self.mu_1)
                * (2 * self.mu_0 * self.mu_2 - self.mu_2**2)
            )
            if (self.mu_0 * self.mu_1) > 0
            else 0
        )

    @cached_property
    def mu3_conc(self) -> NDArrayType:
        """Dead polymer moment 3 concentration"""
        return self.mu_3 / (self.V + EPS)

    def x(self, s0: "ModelState") -> NDArrayType:
        """Conversion of monomer"""
        return 1 - self.M / (s0.M + EPS)

    @cached_property
    def NACL(self) -> NDArrayType:
        """Number average chain length"""
        return self.mu_1 / (self.mu_0 + EPS)

    @cached_property
    def WACL(self) -> NDArrayType:
        """Weight average chain length"""
        return self.mu_2 / (self.mu_1 + EPS)

    @cached_property
    def PDI(self) -> NDArrayType:
        """Dispersity"""
        return self.WACL / (self.NACL + EPS)


class Model(SupportsDeterministicODE):

    name = "FRP1"
    RateClass = RateCoefficients
    StateClass = ModelState

    def ode_function(self):
        from .equations import kinetic_odes

        return kinetic_odes



# React/models/state.py
from abc import ABC
from dataclasses import dataclass, fields, field
from typing import Generic, Type, TypeVar, List, Optional, Callable, Any
from functools import cached_property

import numpy as np

from sparks.core import NDArrayType


def state_field(default: float = 0.0):
    return field(default_factory=lambda: np.array(default))


StateType = TypeVar("StateType", bound="State")


@dataclass
class State(ABC):

    t: NDArrayType = state_field()

    def __post_init__(self):

        # Get max length from all fields
        # max_len = max(
        #     (np.asarray(getattr(self, f.name)).shape[0] for f in fields(self)),
        #     default=1,
        # )

        for name in self.get_species_names():
            value = np.array(getattr(self, name))
            setattr(self, name, np.array(value))

    @classmethod
    def get_species_names(cls) -> List[str]:
        """Get all species field names (excluding time)"""
        # Cache the result to avoid repeated fields() calls
        if not hasattr(cls, "_cached_species_names"):
            cls._cached_species_names = [f.name for f in fields(cls) if f.name != "t"]
        return cls._cached_species_names

    @property
    def is_trajectory(self) -> bool:
        """Check if the state is a trajectory"""
        return self.t.ndim > 0 and len(self.t) > 1

    def to_array(self) -> NDArrayType:
        """Convert state to numpy array (excluding time)"""
        return np.array([getattr(self, name) for name in self.get_species_names()])

    @classmethod
    def from_array(
        cls: Type[StateType],
        arr: NDArrayType,
        t: Optional[NDArrayType] = None,
    ) -> StateType:
        """Create a state from a numpy array"""
        arr = np.asarray(arr)
        species_names = cls.get_species_names()

        if t is None:
            t = np.array(0.0) if arr.ndim == 1 else np.zeros(arr.shape[-1])

        kwargs = {"t": t}
        if arr.ndim == 1:
            for i, name in enumerate(species_names):
                kwargs[name] = arr[i]  # Skip np.array(), let __post_init__ handle it
        else:
            # Trajectory
            for i, name in enumerate(species_names):
                kwargs[name] = arr[i]

        return cls(**kwargs)

    def get_init_state(self) -> "State":
        return self.from_array(self.to_array()[:, 0], t=np.array(0.0))


@dataclass
class SimulationData(Generic[StateType]):
    trajectory: StateType
    init_state: StateType

    @staticmethod
    def from_trajectory(traj: StateType) -> "SimulationData":
        init_state = traj.get_init_state()
        return SimulationData(trajectory=traj, init_state=init_state)


def interpolate_at_time(traj: StateType, target_times: NDArrayType) -> StateType:

    t = traj.t
    target_times = np.asarray(target_times)

    if np.any(target_times < t[0]) or np.any(target_times > t[-1]):
        raise ValueError("Target times are out of bounds.")

    species_names = traj.get_species_names()
    kwargs = {"t": target_times}

    for name in species_names:
        species_values = getattr(traj, name)
        interp_values = np.interp(target_times, t, species_values)
        kwargs[name] = interp_values

    return type(traj)(**kwargs)


def interpolate_by(traj: StateType, x: NDArrayType, xp: NDArrayType) -> StateType:

    x = np.asarray(x)
    xp = np.asarray(xp)

    t_at_x = np.interp(x, xp, traj.t)
    return interpolate_at_time(traj, t_at_x)


def get_midpoint_values(traj: StateType) -> StateType:

    t = traj.t
    t_mid = 0.5 * (t[:-1] + t[1:])

    return interpolate_at_time(traj, t_mid)


def derived(fn: Callable[..., Any] = None, *, cache: bool = True):
    """
    Use on implementation of State to make a property (optionally cached) that
    is also visible to the RHS view builder as a derived relationship.
    """

    if fn is None:
        # allow @derived_property() with kwargs
        return lambda f: derived(f, cache=cache)

    prop = cached_property(fn) if cache else property(fn)
    setattr(prop, "__is_derived__", True)
    return prop



# sparks/core/__init__.py
from typing import TypeAlias
import numpy as np
from numpy.typing import NDArray


NDArrayType: TypeAlias = NDArray[np.floating]

EPS = 1e-40  # Small value to avoid division by zero
NA = 6.022e23  # Avogadro's number (mol^-1)


def okprint(print_str: str):
    """Prints a message in green."""
    print(f"\033[92m{print_str}\033[0m")


def errprint(print_str: str):
    """Prints error message in red."""
    print(f"\033[91m{print_str}\033[0m")


def warnprint(print_str: str):
    """Prints warning message in pale orange."""
    # ANSI 38;5;215 is a pale orange (xterm 256 color)
    print(f"\033[38;5;215m{print_str}\033[0m")


def dbgprint(print_str: str):
    """Prints debug message in light blue."""
    print(f"\033[96m{print_str}\033[0m")


__all__ = [
    "NDArrayType",
    "EPS",
    "NA",
    "okprint",
    "errprint",
    "warnprint",
    "dbgprint",
]



from typing import Any, Dict, List, Tuple

import numpy as np
import pandas as pd
from runkmc.simulation import SimulationResult

from sparks.models import SimulationData
from sparks.simulators.stochastic import KMCConfig
from .model import ModelState, RateCoefficients, ChainModelState, SequenceModelState


def get_FRP2_kmc_inputs(
    k: RateCoefficients, s0: ModelState, config: KMCConfig
) -> Dict[str, Any]:

    return {
        "num_units": config.num_particles,
        "termination_time": config.termination_time,
        "analysis_time": config.analysis_time,
        "I_c0": float(s0.I),
        "R_c0": float(s0.R),
        "A_c0": float(s0.A),
        "B_c0": float(s0.B),
        "I_FW": 0.0,
        "R_FW": 1.0,
        "A_FW": 100.0,
        "B_FW": 200.0,
        "kd": k.kd,
        "kpAA": k.kpAA,
        "kpAB": k.kpAB,
        "kpBA": k.kpBA,
        "kpBB": k.kpBB,
        "kdAA": k.kdAA,
        "kdAB": k.kdAB,
        "kdBA": k.kdBA,
        "kdBB": k.kdBB,
        "ktcAA": k.ktcAA,
        "ktcAB": k.ktcAB,
        "ktcBB": k.ktcBB,
        "ktdAA": k.ktdAA,
        "ktdAB": k.ktdAB,
        "ktdBB": k.ktdBB,
    }


def get_CRP3_kmc_inputs(
    k: RateCoefficients, s0: ModelState, config: KMCConfig
) -> Dict[str, Any]:

    return {
        "num_units": config.num_particles,
        "termination_time": config.termination_time,
        "analysis_time": config.analysis_time,
        "I_c0": float(s0.R),
        "A_c0": float(s0.A),
        "B_c0": float(s0.B),
        "I_FW": 1.0,
        "A_FW": 100.0,
        "B_FW": 200.0,
        "kpAA": k.kpAA,
        "kpAB": k.kpAB,
        "kpBA": k.kpBA,
        "kpBB": k.kpBB,
        "kdAA": k.kdAA,
        "kdAB": k.kdAB,
        "kdBA": k.kdBA,
        "kdBB": k.kdBB,
    }


def get_kmc_inputs(
    k: RateCoefficients, s0: ModelState, config: KMCConfig, model_name: str
) -> Dict[str, Any]:

    if model_name == "CRP3":
        return get_CRP3_kmc_inputs(k, s0, config)
    elif model_name == "FRP2":
        return get_FRP2_kmc_inputs(k, s0, config)
    else:
        raise ValueError(f"Unknown model name: {model_name}")


def build_chain_model_state(df: pd.DataFrame) -> SimulationData[ChainModelState]:

    t = df["KMC Time"].values
    zeros = np.zeros_like(t)
    NACL = df["nAvgChainLength"].values
    WACL = df["wAvgChainLength"].values

    cms = ChainModelState(
        t=t,
        lam_0=zeros,
        lam_1=zeros,
        lam_2=zeros,
        mu_0=np.ones_like(t),
        mu_1=NACL,
        mu_2=WACL * NACL,
    )
    return SimulationData[ChainModelState].from_trajectory(cms)


def build_sequence_model_state(df: pd.DataFrame) -> SimulationData[SequenceModelState]:

    t = df["KMC Time"].values
    zeros = np.zeros_like(t)
    NASL_A = df["nAvgSequenceLength_A"].values
    NASL_B = df["nAvgSequenceLength_B"].values
    WASL_A = df["wAvgSequenceLength_A"].values
    WASL_B = df["wAvgSequenceLength_B"].values

    sms = SequenceModelState(
        t=t,
        aSA0=zeros,
        aSB0=zeros,
        aSA1=zeros,
        aSB1=zeros,
        aSA2=zeros,
        aSB2=zeros,
        iSA0=np.ones_like(t),
        iSB0=np.ones_like(t),
        iSA1=NASL_A,
        iSB1=NASL_B,
        iSA2=WASL_A * NASL_A,
        iSB2=WASL_B * NASL_B,
    )

    return SimulationData[SequenceModelState].from_trajectory(sms)


def build_pos_sequence_model_state(
    df: pd.DataFrame,
) -> List[SimulationData[SequenceModelState]]:

    pos_sms = []
    for pos, group_df in df.groupby("Bucket"):

        group_df = group_df.sort_values("KMC Time")

        t = np.insert(group_df["KMC Time"].values, 0, 0)
        zeros = np.zeros_like(t)
        mc_A = np.insert(group_df["MonomerCount_A"].values, 0, 0)
        mc_B = np.insert(group_df["MonomerCount_B"].values, 0, 0)
        sc_A = np.insert(group_df["SequenceCount_A"].values, 0, 0)
        sc_B = np.insert(group_df["SequenceCount_B"].values, 0, 0)
        sl2_A = np.insert(group_df["SequenceLengths2_A"].values, 0, 0)
        sl2_B = np.insert(group_df["SequenceLengths2_B"].values, 0, 0)

        NASL_A = np.divide(mc_A, sc_A, where=sc_A != 0)
        NASL_B = np.divide(mc_B, sc_B, where=sc_B != 0)
        WASL_A = np.divide(sl2_A, sc_A, where=sc_A != 0)
        WASL_B = np.divide(sl2_B, sc_B, where=sc_B != 0)

        sms = SequenceModelState(
            t=t,
            aSA0=zeros,
            aSB0=zeros,
            aSA1=zeros,
            aSB1=zeros,
            aSA2=zeros,
            aSB2=zeros,
            iSA0=np.ones_like(t),
            iSB0=np.ones_like(t),
            iSA1=NASL_A,
            iSB1=NASL_B,
            iSA2=WASL_A * NASL_A,
            iSB2=WASL_B * NASL_B,
        )
        sms_data = SimulationData[SequenceModelState].from_trajectory(sms)
        pos_sms.append(sms_data)

    return pos_sms


def build_FRP2_model_state(df: pd.DataFrame) -> SimulationData[ModelState]:

    NAV = df["NAV"].values
    t = df["KMC Time"].values
    I = df["I Count"].values / NAV
    R = df["R Count"].values / NAV
    A = df["A Count"].values / NAV
    B = df["B Count"].values / NAV
    RA = df["P[R.A] Count"].values / NAV
    RB = df["P[R.B] Count"].values / NAV
    PAA = df["P[A.A] Count"].values / NAV
    PAB = df["P[A.B] Count"].values / NAV
    PBA = df["P[B.A] Count"].values / NAV
    PBB = df["P[B.B] Count"].values / NAV
    P = PAA + PAB + PBA + PBB
    PD = df["D Count"].values / NAV

    state = ModelState(t, I, R, A, B, RA, RB, PAA, PAB, PBA, PBB, PD)
    init_state = ModelState(
        t=t[0],
        I=I[0],
        R=R[0],
        A=A[0],
        B=B[0],
        RA=RA[0],
        RB=RB[0],
        PAA=PAA[0],
        PAB=PAB[0],
        PBA=PBA[0],
        PBB=PBB[0],
        PD=PD[0],
    )
    return SimulationData[ModelState].from_trajectory(state)
    # return SimulationData[ModelState](trajectory=state, init_state=init_state)


def build_CRP3_model_state(df: pd.DataFrame) -> SimulationData[ModelState]:

    NAV = df["NAV"].values
    t = df["KMC Time"].values
    I = np.zeros_like(t)
    R = df["I Count"].values / NAV
    A = df["A Count"].values / NAV
    B = df["B Count"].values / NAV
    RA = df["P[I.A] Count"].values / NAV
    RB = df["P[I.B] Count"].values / NAV
    PAA = df["P[-.A.A] Count"].values / NAV
    PAB = df["P[-.A.B] Count"].values / NAV
    PBA = df["P[-.B.A] Count"].values / NAV
    PBB = df["P[-.B.B] Count"].values / NAV
    PD = df["D Count"].values / NAV

    state = ModelState(t, I, R, A, B, RA, RB, PAA, PAB, PBA, PBB, PD)
    init_state = ModelState(
        t=t[0],
        I=0,
        R=R[0],
        A=A[0],
        B=B[0],
        RA=RA[0],
        RB=RB[0],
        PAA=PAA[0],
        PAB=PAB[0],
        PBA=PBA[0],
        PBB=PBB[0],
        PD=PD[0],
    )
    return SimulationData[ModelState].from_trajectory(state)
    # return SimulationData[ModelState](trajectory=state, init_state=init_state)


def parse_kmc_outputs(
    result: SimulationResult, model_name: str
) -> Tuple[SimulationData[ModelState], Any]:

    df = result.load_results()

    model_data = None
    if model_name == "FRP2":
        model_data = build_FRP2_model_state(df)
    elif model_name == "CRP3":
        model_data = build_CRP3_model_state(df)
    else:
        raise ValueError(f"Unknown model name: {model_name}")

    cms = build_chain_model_state(df)
    sms = build_sequence_model_state(df)

    seq_df = result.load_sequence_data()
    pos_sms = None
    if seq_df is not None and not seq_df.empty:
        pos_sms = build_pos_sequence_model_state(seq_df)

    mdata = {"cms": cms, "sms": sms, "pos_sms": pos_sms}

    return model_data, mdata



def okprint(print_str: str):
    """Prints a message in green."""
    print(f"\033[92m{print_str}\033[0m")


def errprint(print_str: str):
    """Prints error message in red."""
    print(f"\033[91m{print_str}\033[0m")


def warnprint(print_str: str):
    """Prints warning message in pale orange."""
    # ANSI 38;5;215 is a pale orange (xterm 256 color)
    print(f"\033[38;5;215m{print_str}\033[0m")


def dbgprint(print_str: str):
    """Prints debug message in light blue."""
    print(f"\033[96m{print_str}\033[0m")



from enum import Enum
from dataclasses import dataclass
from typing import Dict

from sparks.utils.formulas import arrhenius_eqn, kcal_to_J, get_arrhenius_const, C_to_K
from sparks.utils.constants import R_cal
from .monomers import MonomerType


class InitiatorType(Enum):
    General = "General"
    AIBN = "AIBN"
    BPO = "BPO"
    dTBPO = "dTBPO"
    TBPA = "TBPA"
    TBPB = "TBPB"


@dataclass
class InitiatorData:

    name: str
    id: InitiatorType
    # id: str

    # Initiator molecular weight (g/mol)
    MW: float

    # Decomposition rate constant (L/min)
    kd: Dict[MonomerType, float]

    # Initiator efficiency
    f: Dict[MonomerType, float]

    # Critical free volume for diffusion-control
    Vf_i: Dict[MonomerType, float]

    # Rate of decrease of f
    C: Dict[MonomerType, float]


def AIBN(T_K: float) -> InitiatorData:
    return InitiatorData(
        name="AIBN",
        id=InitiatorType.AIBN,
        MW=164.21,
        kd={
            MonomerType.Sty: arrhenius_eqn(6.33e16, 3.0719e4, R=R_cal)(T_K),
            MonomerType.EA: arrhenius_eqn(7.7803e16, 3.0704e4, R=R_cal)(T_K),
            MonomerType.General: arrhenius_eqn(6.23e16, 3.0704e4, R=R_cal)(T_K),
        },
        f={
            MonomerType.Sty: 0.6,
            MonomerType.BMA: 0.42,
            MonomerType.General: arrhenius_eqn(0.0247, -2166, R=R_cal)(T_K),
        },
        Vf_i={
            MonomerType.Sty: 0.04,
            MonomerType.BA: 0.15,
            MonomerType.EA: arrhenius_eqn(0.825, 1175, R=R_cal)(T_K),
            MonomerType.BMA: 0.09,
            MonomerType.General: arrhenius_eqn(0.6365, 1368.8, R=R_cal)(T_K),
        },
        C={
            MonomerType.Sty: 0.5,
            MonomerType.BA: 1,
            MonomerType.EA: 1,
            MonomerType.General: 0.685,
        },
    )


def dTBPO(T_K: float) -> InitiatorData:
    return InitiatorData(
        name="di-tert-butyl peroxide",
        id=InitiatorType.dTBPO,
        MW=146.23,
        kd={
            # MonomerType.MMA: arrhenius_eqn(1.68e16, 3.5e4, R=R_cal)(T_K),
            MonomerType.MMA: arrhenius_eqn(7.29e16, 3.56e4, R=R_cal)(T_K),
            MonomerType.General: arrhenius_eqn(7.29e16, 3.56e4, R=R_cal)(T_K),
        },
        f={
            MonomerType.MMA: 0.7,
            MonomerType.BA: 0.3,
            MonomerType.General: 0.5,
        },
        Vf_i={MonomerType.General: 0.15},
        C={MonomerType.General: 0.25},
    )



from dataclasses import dataclass

import numpy as np

from sparks.core import NDArrayType, okprint, errprint, warnprint
from sparks.data import MonomerType

from .model import RateCoefficients, ModelState, SequenceModelState


def get_kt_from_Cheong_2003(s: ModelState, k: RateCoefficients):

    T_C = k.T_C
    if T_C == 115:
        ka = 0.23e14  # L/mol/min
    elif T_C == 140:
        ka = 0.8e14  # L/mol/min
    else:
        raise ValueError(f"Unsupported temperature: {T_C} C")

    ktBB = k.mpropsB.kt
    ktAA = (np.sqrt(ka) - s.beta * np.sqrt(ktBB)) ** 2
    ktAB = np.sqrt(ktAA * ktBB)

    return ktAA, ktAB, ktBB


@dataclass
class Context:

    # Parameters
    use_diffusion_control: bool = False
    const_volume: bool = False
    const_temperature: bool = True

    # Initial state
    init_state: ModelState | None = None

    # Gel Effect
    triggered: bool = False
    t_event: NDArrayType | None = None
    Vf_cr1: float | None = None
    kt_cr: float | None = None
    Mw_cr: float | None = None

    debug_counter: int = 0
    debug_interval: int = 100

    # times = [10, 20, 30, 40, 50, 60, 70, 80, 90]
    times = [100, 200, 300, 400, 500, 600, 700, 800, 900]
    # times = []
    time_idx: int = 0


def kinetic_odes(
    dydt: ModelState,
    k: RateCoefficients,
    t: NDArrayType,
    s: ModelState,
    context: Context = Context(),
) -> ModelState:
    """Defines the ODEs for the full FRP2 model."""

    dydt.t = t

    T_C = k.T_C
    T_K = T_C + 273.15  # Convert Celsius to Kelvin

    # Extract properties
    iprops = k.iprops
    mpropsA = k.mpropsA
    mpropsB = k.mpropsB

    # Save initial state
    if not context.init_state and t >= 0:
        context.init_state = s

    # A0, B0 are initial moles for volume balance calculations
    A0 = context.init_state.A
    B0 = context.init_state.B
    M0 = A0 + B0

    ######################
    ### Rate Constants ###
    ######################

    kpAA, kpAB, kpBA, kpBB = k.kp()
    kpRAA, kpRAB, kpRBA, kpRBB = k.kp(f=1e1)
    kdAA, kdAB, kdBA, kdBB = k.kd()

    # Initiation
    kd = iprops.kd[MonomerType.General]
    f = iprops.f[MonomerType.General]

    # Thermal Initiation
    kthA = mpropsA.kth
    kthB = mpropsB.kth

    # Chain Transfer to Monomer
    kf_mAA, kf_mAB, kf_mBA, kf_mBB = k.kfm()
    kfm = kfm_pseudo = (
        kf_mAA * s.pA * s.fA
        + kf_mAB * s.pA * s.fB
        + kf_mBA * s.pB * s.fA
        + kf_mBB * s.pB * s.fB
    )

    kf_mAA = 0
    kf_mBB = 0
    kf_mAB = 0
    kf_mBA = 0

    RpA = (
        s.A_conc
        * (
            kpRAA * (s.R_conc + s.RA_conc)
            + kpRBA * s.RB_conc
            + kpAA * s.PA_conc
            + kpBA * s.PB_conc
        )
        - s.PA_conc * (kdAA * s.pAA + kdBA * s.pBA)
        + s.A_conc * (kf_mAA * s.PA_conc + kf_mBA * s.PB_conc)
        + 2 * kthA * s.A_conc**3
    )
    RpB = (
        s.B_conc
        * (
            kpRBB * (s.R_conc + s.RB_conc)
            + kpRAB * s.RA_conc
            + kpBB * s.PB_conc
            + kpAB * s.PA_conc
        )
        - s.PB_conc * (kdBB * s.pBB + kdAB * s.pAB)
        + s.B_conc * (kf_mAB * s.PA_conc + kf_mBB * s.PB_conc)
        + 2 * kthB * s.B_conc**3
    )
    Rp = RpA + RpB

    # Calculate composition of the polymer
    F_A = RpA / Rp if Rp > 0 else 0.0
    F_B = RpB / Rp if Rp > 0 else 0.0

    ###########################################
    ### Calculate the pseudo rate constants ###
    ###########################################

    # Pseudo rate constant for propagation
    # Mistake in Eq. 52? kdij * Pij * pi / M_conc
    # Should be kdij * Pij * pj / M_conc
    kp = kp_pseudo = (
        # i=A, j=A
        (kpAA * s.pA * s.fA - kdAA * s.pAA * s.pA / s.M_conc)
        # i=A, j=B
        + (kpAB * s.pA * s.fB - kdAB * s.pAB * s.pB / s.M_conc)
        # i=B, j=A
        + (kpBA * s.pB * s.fA - kdBA * s.pBA * s.pA / s.M_conc)
        # i=B, j=B
        + (kpBB * s.pB * s.fB - kdBB * s.pBB * s.pB / s.M_conc)
    )

    ktAA, ktAB, ktBB = get_kt_from_Cheong_2003(s, k)
    kt = kt_pseduo = ktAA * s.pA * s.pA + ktBB * s.pB * s.pB + 2 * ktAB * s.pA * s.pB
    ktd_ratio = mpropsA.ktd_ratio * s.pA + mpropsB.ktd_ratio * s.pB

    # Transfer to impurity, solvent, CTA, or polymer not implemented
    kfz = 0
    kfs = 0
    kfCTA = 0
    kfp = 0

    ########################################
    ### Calculate the overall properties ###
    ########################################

    # Effective molecular weight of monomer/polymer
    MW_eff = s.FbarA * mpropsA.MW + s.FbarB * mpropsB.MW

    # Calculate the mass/volume of monomer
    m_A = s.A * mpropsA.MW
    m_B = s.B * mpropsB.MW
    vol_A = m_A / mpropsA.dens_m
    vol_B = m_B / mpropsB.dens_m
    vol_m = vol_A + vol_B

    # Calculate the mass/volume of polymer
    m_PA = s.nPA * mpropsA.MW
    m_PB = s.nPB * mpropsB.MW
    m_P = m_PA + m_PB
    vol_PA = m_PA / mpropsA.dens_p
    vol_PB = m_PB / mpropsB.dens_p
    vol_P = vol_PA + vol_PB  # Volume of polymer in L

    kt_factor = 1
    kp_factor = 1
    kfm_factor = 1

    if context.use_diffusion_control:

        Mn = MW_eff * s.NACL
        Mw = MW_eff * s.WACL

        Vf_p = mpropsA.Vf_p
        alpha_mP = mpropsA.alpha_p

        wA = m_PA / m_P if m_P > 0 else 0
        wB = m_PB / m_P if m_P > 0 else 0
        Tg_P = 1 / (wA / mpropsA.Tg_p + wB / mpropsB.Tg_p) if m_P > 0 else 0

        # Calculate the free volume
        Vf_A = (mpropsA.Vf_m + mpropsA.alpha_m * (T_K - mpropsA.Tg_m)) * vol_A / s.V
        Vf_B = (mpropsB.Vf_m + mpropsB.alpha_m * (T_K - mpropsB.Tg_m)) * vol_B / s.V
        Vf_P = (Vf_p + alpha_mP * (T_K - Tg_P)) * vol_P / s.V
        Vf = Vf_A + Vf_B + Vf_P

        delta_c = mpropsA.delta * s.pA + mpropsB.delta * s.pB

        K3 = K3_pseudo = np.exp(
            1 / (s.FbarA / np.log(mpropsA.K3) + s.FbarB / np.log(mpropsB.K3))
        )

        A_factor = A_pseudo = (
            1 / (F_A / mpropsA.A + F_B / mpropsB.A) if (F_A + F_B) > 0 else 0
        )
        m = np.where((s.pA + s.pB) > 0, mpropsA.m * s.pA + mpropsB.m * s.pB, 1)
        n = np.where((s.pA + s.pB) > 0, mpropsA.n * s.pA + mpropsB.n * s.pB, 1)

        if not context.triggered:
            # 1) Segmental diffusion termination

            c = m_P / s.V  # Mass concentration of accumulated polymer (g/L)
            # kt = kt * (1 + delta_c * c)

            K3_test = (Mw**m) * np.exp(A_factor / Vf)
            # print(
            #     f"K3_test={K3_test:.2f}, K3={K3:.2f}, K3_A={mpropsA.K3:.2f}, K3_B={mpropsB.K3:.2f}"
            # )
            # warnprint(K3_test)
            # warnprint(K3)
            # warnprint(f"Mw={Mw:.3e} g/mol")
            if not np.isclose(K3, 0) and K3_test >= K3 and t >= 1:
                context.triggered = True
                context.t_event = t
                context.Vf_cr1 = Vf * 1.0
                context.kt_cr = kt
                context.Mw_cr = Mw

                warnprint(
                    f"K3 triggered at t={t:.2f} s: K3_test={K3_test:.3e}, K3={K3:.3e}"
                )
                warnprint(
                    f"A (A factor): {mpropsA.A}, B (A factor): {mpropsB.A}, A_factor: {A_factor}"
                )
                warnprint(f"A (K3): {mpropsA.K3}, B (K3): {mpropsB.K3}, K3: {K3}")
                warnprint(
                    f"F_A: {F_A}, F_B: {F_B}, Fbar_A: {s.FbarA}, Fbar_B: {s.FbarB}"
                )
                warnprint(
                    f"Triggered at t={t:.2f} s, Vf={Vf:.2f} L, Mw={Mw/1e3:.2f} kDa"
                )
        else:

            kt_factor = (context.Mw_cr / Mw) ** n * np.exp(
                -A_factor * (1 / Vf - 1 / context.Vf_cr1)
            )

        # 2) Translational diffusion termination
        Vf_cr2 = mpropsA.Vf_c * s.pA + mpropsB.Vf_c * s.pB
        B_factor = mpropsA.B * s.pA + mpropsB.B * s.pB
        if Vf < Vf_cr2:
            kp_factor = kfm_factor = np.exp(-B_factor * (1 / Vf - 1 / Vf_cr2))
            # kp = kp * np.exp(-B_factor * (1 / Vf - 1 / Vf_cr2))
            # kfm = kfm * np.exp(-B_factor * (1 / Vf - 1 / Vf_cr2))

        Vf_i = iprops.Vf_i[MonomerType.General]
        C_factor = iprops.C[MonomerType.General]
        if Vf < Vf_i:
            f = f * np.exp(-C_factor * (1 / Vf - 1 / Vf_i))

        # 3) Reaction diffusion termination
        # D = (
        #     mprops.n_s * (mprops.l0) ** 2 * kp * M_conc / 6.0
        # )  # Diffusion coefficient (m^2/s)
        # delta = (6 * (vol_m / nM) / (np.pi * NA)) ** (1 / 3.0)
        # # delta = mprops.delta
        # kt_rd = 8 * np.pi * NA * delta * D / 1000
        # kt += 0.0 * kt_rd

    ktAA *= kt_factor
    ktBB *= kt_factor
    ktAB *= kt_factor
    kt *= kt_factor

    kpAA, kpAB, kpBA, kpBB = k.kp(f=kp_factor)
    kdAA, kdAB, kdBA, kdBB = k.kd(f=kp_factor)
    kp *= kp_factor

    kf_mAA, kf_mAB, kf_mBA, kf_mBB = k.kfm(f=kfm_factor)
    kfm_pseudo *= kfm_factor

    ktc = kt / (1 + ktd_ratio)
    ktd = kt * ktd_ratio / (1 + ktd_ratio)
    # ktc = kt
    # ktd = 0
    # ktc = 0
    # ktd = kt

    Ri = 2 * f * kd * s.I_conc + 2 * kthA * s.A_conc**3 + 2 * kthB * s.B_conc**3

    # Rates of Propagation of A and B
    RpA = (
        s.A_conc
        * (
            kpRAA * (s.R_conc + s.RA_conc)
            + kpRBA * s.RB_conc
            + kpAA * s.PA_conc
            + kpBA * s.PB_conc
        )
        - s.PA_conc * (kdAA * s.pAA + kdBA * s.pBA)
        + s.A_conc * (kf_mAA * s.PA_conc + kf_mBA * s.PB_conc)
        + 2 * kthA * s.A_conc**3
    )
    RpB = (
        s.B_conc
        * (
            kpRBB * (s.R_conc + s.RB_conc)
            + kpRAB * s.RA_conc
            + kpBB * s.PB_conc
            + kpAB * s.PA_conc
        )
        - s.PB_conc * (kdBB * s.pBB + kdAB * s.pAB)
        + s.B_conc * (kf_mAB * s.PA_conc + kf_mBB * s.PB_conc)
        + 2 * kthB * s.B_conc**3
    )
    Rp = RpA + RpB

    # Calculate volume change from monomer
    vdelta_mA = RpA * vol_A / s.A if s.A > 0 else 0.0
    vdelta_mB = RpB * vol_B / s.B if s.B > 0 else 0.0
    vdelta_PA = RpA * vol_PA / s.nPA if s.nPA > 0 else 0.0
    vdelta_PB = RpB * vol_PB / s.nPB if s.nPB > 0 else 0.0
    vdelta = vdelta_mA + vdelta_mB - vdelta_PA - vdelta_PB

    if context.const_volume:
        dydt.V = 0 * s.V
    else:
        dydt.V = -s.V * vdelta

    """ Moles of Initiator """
    dydt.I = s.V * (-kd * s.I_conc)

    """ Moles of Free Radical, R """
    dydt.R = s.V * (
        # Formation by initiation
        +2 * f * kd * s.I_conc
        # Consumption by propagation
        - kpRAA * s.R_conc * s.A_conc
        - kpRBB * s.R_conc * s.B_conc
    )

    """ Moles of A Monomer """
    dydt.A = s.V * (
        # Consumption by propagation
        -kpRAA * (s.R_conc + s.RA_conc) * s.A_conc
        - kpRBA * s.RB_conc * s.A_conc
        - kpAA * s.PA_conc * s.A_conc
        - kpBA * s.PB_conc * s.A_conc
        # Formation by depropagation
        + kdAA * s.PAA_conc
        + kdBA * s.PBA_conc
        # Consumption by chain transfer to monomer
        - kf_mAA * s.PA_conc * s.A_conc
        - kf_mBA * s.PB_conc * s.A_conc
        # Consumption by thermal initiation
        - 3 * kthA * s.A_conc**3
    )

    """ Moles of B Monomer """
    dydt.B = s.V * (
        # Consumption by propagation
        -kpRBB * (s.R_conc + s.RB_conc) * s.B_conc
        - kpRAB * s.RA_conc * s.B_conc
        - kpAB * s.PA_conc * s.B_conc
        - kpBB * s.PB_conc * s.B_conc
        # Formation by depropagation
        + kdAB * s.PAB_conc
        + kdBB * s.PBB_conc
        # Consumption by chain transfer to monomer
        - kf_mAB * s.PA_conc * s.B_conc
        - kf_mBB * s.PB_conc * s.B_conc
        # Consumption by thermal initiation
        - 3 * kthB * s.B_conc**3
    )

    """ Moles of R-A dyad"""
    dydt.RA = s.V * (
        # Formation by thermal initiation (3A -> 2RA)
        +2 * kthA * s.A_conc**3
        # Formation by propagation (R + A -> RA)
        + kpRAA * s.R_conc * s.A_conc
        # Consumption by propagation (RA + A -> PAA, RA + B -> PAB)
        - kpRAA * s.RA_conc * s.A_conc
        - kpRAB * s.RA_conc * s.B_conc
        # Formation by chain transfer (PA + A -> PD + RA)
        + kf_mAA * s.PA_conc * s.A_conc
        # Formation by chain transfer (PB + A -> PD + RA)
        + kf_mBA * s.PB_conc * s.A_conc
    )

    """ Total moles of R-B dyad """
    dydt.RB = s.V * (
        # Formation by thermal initiation (3B -> 2RB)
        +2 * kthB * s.B_conc**3
        # Formation by propagation (R + B -> RB)
        + kpRBB * s.R_conc * s.B_conc
        # Consumption by propagation (RB + A -> PBA, RB + B -> PBB)
        - kpRBA * s.RB_conc * s.A_conc
        - kpRBB * s.RB_conc * s.B_conc
        # Formation by chain transfer (PA + B -> PD + RB)
        + kf_mAB * s.PA_conc * s.B_conc
        # Formation by chain transfer (PB + B -> PD + RB)
        + kf_mBB * s.PB_conc * s.B_conc
    )

    c_tAA = c_tBB = 1
    c_tAB = 1

    """ Total moles of ~~AA* polymer """
    dydt.PAA = s.V * (
        # Formation by propagation (RA + A -> PAA)
        +kpRAA * s.RA_conc * s.A_conc
        # Formation by propagation (PBA + A -> PAA)
        + kpAA * s.PBA_conc * s.A_conc
        # Consumption by propagation (PAA + B -> PAB)
        - kpAB * s.PAA_conc * s.B_conc
        # Formation by depropagation (PAAB -> PAA + B)
        + kdAB * s.pAA * s.PAB_conc
        # Consumption by depropagation (PBAA -> PBA + A)
        - kdAA * s.pBA * s.PAA_conc
        # Consumption by chain transfer to monomer (PAA + B -tr-> PD + RB)
        - kf_mAB * s.PAA_conc * s.B_conc
        # Consumption by chain transfer to monomer (PAA + A -tr-> PD + RA)
        - kf_mAA * s.PAA_conc * s.A_conc
        # Consumption by termination (PAA + PA -> (2)PD, PAA + PB -> (2)PD)
        - c_tAA * ktAA * s.PAA_conc * s.PA_conc
        - c_tAB * ktAB * s.PAA_conc * s.PB_conc
    )

    """ Total moles of ~~AB* polymer """
    dydt.PAB = s.V * (
        # Formation by propagation (RA + B -> PAB)
        +kpRAB * s.RA_conc * s.B_conc
        # Formation by propagation (PA + B -> PAB)
        + kpAB * s.PA_conc * s.B_conc
        # Consumption by propagation (PAB + A -> PBA)
        - kpBA * s.PAB_conc * s.A_conc
        # Consumption by propagation (PAB + B -> PBB)
        - kpBB * s.PAB_conc * s.B_conc
        # Formation by depropagation (PABA -> PAB + A)
        + kdBA * s.pAB * s.PBA_conc
        # Formation by depropagation (PABB -> PAB + B)
        + kdBB * s.pAB * s.PBB_conc
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.PAB_conc
        # Consumption by chain transfer to monomer (PAB + A -tr-> PD + RA)
        - kf_mBA * s.PAB_conc * s.A_conc
        # Consumption by chain transfer to monomer (PAB + B -tr-> PD + RB)
        - kf_mBB * s.PAB_conc * s.B_conc
        # Consumption by termination (PAB + PA -> (2)PD, PAB + PB -> (2)PD)
        - c_tAB * ktAB * s.PAB_conc * s.PA_conc
        - c_tBB * ktBB * s.PAB_conc * s.PB_conc
    )

    """ Total moles of ~~BA* polymer """
    dydt.PBA = s.V * (
        # Formation by propagation (RB + A -> PBA)
        +kpRBA * s.RB_conc * s.A_conc
        # Formation by propagation (PB + A -> PBA)
        + kpBA * s.PB_conc * s.A_conc
        # Consumption by propagation (PBA + A -> PAA)
        - kpAA * s.PBA_conc * s.A_conc
        # Consumption by propagation (PBA + B -> PBB)
        - kpAB * s.PBA_conc * s.B_conc
        # Formation by depropagation (PBAA -> PBA + A)
        + kdAA * s.pBA * s.PAA_conc
        # Formation by depropagation (PBAB -> PBA + B)
        + kdAB * s.pBA * s.PAB_conc
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.PBA_conc
        # Consumption by chain transfer to monomer (PBA + A -tr-> PD + RA)
        - kf_mAA * s.PBA_conc * s.A_conc
        # Consumption by chain transfer to monomer (PBA + B -tr-> PD + RB)
        - kf_mAB * s.PBA_conc * s.B_conc
        # Consumption by termination (PBA + PA -> (2)PD, PBA + PB -> (2)PD)
        - c_tAA * ktAA * s.PBA_conc * s.PA_conc
        - c_tAB * ktAB * s.PBA_conc * s.PB_conc
    )

    """ Total moles of ~~BB* polymer """
    dydt.PBB = s.V * (
        # Formation by propagation (RB + B -> PBB)
        +kpRBB * s.RB_conc * s.B_conc
        # Formation by propagation (PAB + B -> PBB)
        + kpBB * s.PAB_conc * s.B_conc
        # Consumption by propagation (PBB + A -> PBA)
        - kpBA * s.PBB_conc * s.A_conc
        # Formation by depropagation (PBBA -> PBB + A)
        + kdBA * s.pBB * s.PBA_conc
        # Consumption by depropagation (PABB -> PAB + B)
        - kdBB * s.pAB * s.PBB_conc
        # Consumption by chain transfer to monomer (PBB + A -tr-> PD + RA)
        - kf_mBA * s.PBB_conc * s.A_conc
        # Consumption by chain transfer to monomer (PBB + B -tr-> PD + RB)
        - kf_mBB * s.PBB_conc * s.B_conc
        # Consumption by termination (PBB + PA -> (2)PD, PBB + PB -> (2)PD)
        - c_tBB * ktBB * s.PBB_conc * s.PB_conc
        - c_tAB * ktAB * s.PBB_conc * s.PA_conc
    )

    """ Total moles of A and B monomer in polymer"""
    dydt.nPA = s.V * RpA
    dydt.nPB = s.V * RpB

    # Formation of polymer from active radicals
    poly_form = (
        kpRAA * s.R_conc * s.A_conc
        + kpRBB * s.R_conc * s.B_conc
        + 2 * kthA * s.A_conc**3
        + 2 * kthB * s.B_conc**3
    )

    mu_3_conc = np.where(
        (s.mu_0_conc * s.mu_1_conc) > 0,
        s.mu_2_conc
        / (s.mu_0_conc * s.mu_1_conc)
        * (2 * s.mu_0_conc * s.mu_2_conc - s.mu_2_conc),
        0,
    )  # Could be misprint (potentially s.mu_1^2)

    """ Active radicals (0th order moment) """
    dydt.lam_0 = s.V * (
        # Formation by propagation
        poly_form
        # Consumption by termination
        - kt * s.lam_0_conc**2
        # Consumption from impurity
        - kfz * s.lam_0_conc * s.Z_conc
    )

    """ Active radicals (1st order moment) """
    dydt.lam_1 = s.V * (
        # Formation by propagation
        poly_form
        # Formation by propagation
        + kp * s.M_conc * s.lam_0_conc
        # Consumption by termination
        - kt * s.lam_0_conc * s.lam_1_conc
        # Consumption from chain transfer
        - (kfCTA * s.CTA_conc + kfm * s.M_conc + kfs * s.S_conc)
        * (s.lam_1_conc - s.lam_0_conc)
        # Formation by chain transfer to polymer
        + kfp * (s.lam_0_conc * mu_3_conc - s.lam_1_conc * s.mu_1_conc)
        # Consumption by chain transfer to impurity
        - kfz * s.lam_1_conc * s.Z_conc
    )

    """Active radicals (2nd order moment)"""
    dydt.lam_2 = s.V * (
        # Formation by propagation
        poly_form
        # Formation by propagation
        + kp * s.M_conc * (s.lam_0_conc + 2 * s.lam_1_conc)
        # Consumption by termination
        - kt * s.lam_0_conc * s.lam_2_conc
        # Consumption from chain transfer
        - (kfCTA * s.CTA_conc + kfm * s.M_conc + kfs * s.S_conc)
        * (s.lam_2_conc - s.lam_0_conc)
        # Formation by chain transfer to polymer
        + kfp * (s.lam_0_conc * mu_3_conc - s.lam_2_conc * s.mu_1_conc)
        # Consumption by chain transfer to impurity
        - kfz * s.lam_2_conc * s.Z_conc
    )

    """Dead radicals (0th order moment)"""
    dydt.mu_0 = s.V * (
        # Formation by termination
        +(0.5 * ktc + ktd) * s.lam_0_conc**2
        # Formation by chain transfer
        + (kfm * s.M_conc + kfCTA * s.CTA_conc + kfs * s.S_conc + kfz * s.Z_conc)
        * s.lam_0_conc
    )

    """Dead radicals (1st order moment)"""
    dydt.mu_1 = s.V * (
        # Formation by termination
        +(ktc + ktd) * s.lam_0_conc * s.lam_1_conc
        # Formation by chain transfer
        + (kfm * s.M_conc + kfCTA * s.CTA_conc + kfs * s.S_conc + kfz * s.Z_conc)
        * s.lam_1_conc
        # Chain transfer to polymer
        + kfp * (s.lam_1_conc * s.mu_1_conc - s.lam_0_conc * s.mu_2_conc)
    )

    """Dead radicals (2nd order moment)"""
    dydt.mu_2 = s.V * (
        # Formation by termination (comb.)
        +ktc * (s.lam_0_conc * s.lam_2_conc + s.lam_1_conc**2)
        # Formation by termination (disp.)
        + ktd * s.lam_0_conc * s.lam_2_conc
        # Formation by chain transfer
        + (kfm * s.M_conc + kfCTA * s.CTA_conc + kfs * s.S_conc + kfz * s.Z_conc)
        * s.lam_2_conc
        # Chain transfer to polymer
        + kfp * (s.lam_2_conc * s.mu_1_conc - s.lam_0_conc * mu_3_conc)
    )

    vol_error = np.where(s.V > 0, 100 * abs((vol_m + vol_P - s.V) / s.V), 0)
    mole_error = np.where(M0 > 0, 100 * abs((s.A + s.B + s.nPA + s.nPB - M0) / M0), 0)

    dRtotdt = dydt.R + dydt.RA + dydt.RB + dydt.PAA + dydt.PAB + dydt.PBA + dydt.PBB
    dRtotdt_ = s.V * (Ri - kt * s.P_conc**2)
    dRtotdt_error = np.where(dRtotdt > 0, 100 * abs((dRtotdt - dRtotdt_) / dRtotdt), 0)

    lam_0_error = np.where(
        s.lam_0_conc > 0, 100 * abs((s.lam_0_conc - s.P_conc) / s.lam_0_conc), 0
    )

    time = np.inf
    if context.time_idx < len(context.times):
        time = context.times[context.time_idx]

    if t >= time:
        # if True:
        context.time_idx += 1
        vol_error = 100 * abs((vol_m + vol_P - s.V) / s.V)
        print(f"Step {context.time_idx}: t={t:.2f}")

        print(f"Time: {t:.2f}, Volume: {s.V:.3e} L")
        print(f"kd: {kd:.3e}, kp: {kp:.3e}, kt: {kt:.3e}, kfm: {kfm:.3e}")
        print(
            f"kfm_factor: {kfm_factor:.3e}, kp_factor: {kp_factor:.3e}, kt_factor: {kt_factor:.3e}"
        )
        print(
            f"pA: {s.pA:.3f}, pB: {s.pB:.3f}, pAA: {s.pAA:.3f}, pAB: {s.pAB:.3f}, pBA: {s.pBA:.3f}, pBB: {s.pBB:.3f}"
        )
        print(f"A_conc: {s.A_conc:.3e}, B_conc: {s.B_conc:.3e} ")

        mole_error_str = f"\tMole balance error: {mole_error:.3e}% (moles = {M0:.3e})"
        if mole_error > 1.0:
            errprint(mole_error_str)
        else:
            okprint(mole_error_str)

        vol_error_str = f"\tVolume balance error: {vol_error:.3e}% (vol = {s.V:.3e} L)"
        if vol_error > 1.0:
            errprint(vol_error_str)
        else:
            okprint(vol_error_str)

        dRtotdt_error_str = f"\tRadical balance error: {dRtotdt_error:.3e}%"
        if dRtotdt_error > 1.0:
            errprint(dRtotdt_error_str)
            errprint(f"\t\tCalculated dRtotdt: {dRtotdt:.3e}, Expected: {dRtotdt_:.3e}")
        else:
            okprint(dRtotdt_error_str)

        lam_0_error_str = f"\tlam_0 error: {lam_0_error:.3e}%"
        if lam_0_error > 1.0:
            errprint(lam_0_error_str)
            errprint(
                f"\t\tCalculated lam_0: {s.lam_0_conc:.3e}, Expected: {s.P_conc:.3e}"
            )
        else:
            okprint(lam_0_error_str)

        print(f"\tActive NACL: {s.active_NACL:.2f}")
        print(f"\tInactive NACL: {s.NACL:.2f}")
        print(f"\t mu0: {s.mu_0:.3e} mu1: {s.mu_1:.3e} mu2: {s.mu_2:.3e}")
        print()

    return dydt


def sequence_odes(
    dydt: SequenceModelState,
    k: RateCoefficients,
    t: NDArrayType,
    sms: SequenceModelState,
    s: ModelState,
) -> SequenceModelState:

    dydt.t = t

    kpAA, kpAB, kpBA, kpBB = k.kp()
    kdAA, kdAB, kdBA, kdBB = k.kd()
    ktAA, ktAB, ktBB = get_kt_from_Cheong_2003(s, k)

    c_tAA = c_tBB = 1
    c_tAB = 1

    kf_mAA, kf_mAB, kf_mBA, kf_mBB = k.kfm()

    aSA0_conc = sms.aSA0 / s.V
    aSB0_conc = sms.aSB0 / s.V
    aSA1_conc = sms.aSA1 / s.V
    aSB1_conc = sms.aSB1 / s.V
    aSA2_conc = sms.aSA2 / s.V
    aSB2_conc = sms.aSB2 / s.V

    # print(f"aSA0_conc: {aSA0_conc:.3e}, aSB0_conc: {aSB0_conc:.3e}")
    # print(f"aSA1_conc: {aSA1_conc:.3e}, aSB1_conc: {aSB1_conc:.3e}")
    # print(f"aSA2_conc: {aSA2_conc:.3e}, aSB2_conc: {aSB2_conc:.3e}")

    # Rates of termination and chain transfer to monomer
    # Ctrij_x = c_tij * ktij * sms.aSijx * sms.aSij0 + c_tij * kf_mij * sms.aSijx * s.j
    RtAA_0 = ktAA * aSA0_conc * aSA0_conc + 0 * kf_mAA * aSA0_conc * s.A_conc
    RtAB_0 = ktAB * aSA0_conc * aSB0_conc + 0 * kf_mAB * aSA0_conc * s.B_conc
    RtBA_0 = ktAB * aSB0_conc * aSA0_conc + 0 * kf_mBA * aSB0_conc * s.A_conc
    RtBB_0 = ktBB * aSB0_conc * aSB0_conc + 0 * kf_mBB * aSB0_conc * s.B_conc
    RtAA_1 = ktAA * aSA1_conc * aSA0_conc + 0 * kf_mAA * aSA1_conc * s.A_conc
    RtAB_1 = ktAB * aSA1_conc * aSB0_conc + 0 * kf_mAB * aSA1_conc * s.B_conc
    RtBA_1 = ktAB * aSB1_conc * aSA0_conc + 0 * kf_mBA * aSB1_conc * s.A_conc
    RtBB_1 = ktBB * aSB1_conc * aSB0_conc + 0 * kf_mBB * aSB1_conc * s.B_conc
    RtAA_2 = ktAA * aSA2_conc * aSA0_conc + 0 * kf_mAA * aSA2_conc * s.A_conc
    RtAB_2 = ktAB * aSA2_conc * aSB0_conc + 0 * kf_mAB * aSA2_conc * s.B_conc
    RtBA_2 = ktAB * aSB2_conc * aSA0_conc + 0 * kf_mBA * aSB2_conc * s.A_conc
    RtBB_2 = ktBB * aSB2_conc * aSB0_conc + 0 * kf_mBB * aSB2_conc * s.B_conc

    a10_frac = aSA1_conc / aSA0_conc if aSA0_conc > 0 else 0
    b10_frac = aSB1_conc / aSB0_conc if aSB0_conc > 0 else 0
    a20_frac = aSA2_conc / aSA0_conc if aSA0_conc > 0 else 0
    b20_frac = aSB2_conc / aSB0_conc if aSB0_conc > 0 else 0

    # Active A sequences (0th order moment)
    dydt.aSA0 = s.V * (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +kpAA * s.A_conc * (s.R_conc + aSA0_conc)
        + kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - kpAA * s.A_conc * aSA0_conc
        - kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + kdAA * s.pAA * aSA0_conc
        + kdAB * s.pAB * aSB0_conc
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - kdAA * s.pAA * aSA0_conc
        - kdBA * s.pBA * aSA0_conc
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_0
        - RtAB_0
    )

    # Active B sequences (0th order moment)
    dydt.aSB0 = s.V * (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +kpBB * s.B_conc * (s.R_conc + aSB0_conc)
        + kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - kpBB * s.B_conc * aSB0_conc
        - kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + kdBB * s.pBB * aSB0_conc
        + kdBA * s.pBA * aSA0_conc
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - kdBB * s.pBB * aSB0_conc
        - kdAB * s.pAB * aSB0_conc
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_0
        - RtBA_0
    )

    # Active A sequences (1st order moment)
    dydt.aSA1 = s.V * (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +kpAA * s.A_conc * (s.R_conc + aSA0_conc + aSA1_conc)
        + kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - kpAA * s.A_conc * aSA1_conc
        - kpAB * s.B_conc * (s.RA_conc + aSA1_conc)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + kdAA * s.pAA * (aSA1_conc - aSA0_conc)
        + kdAB * s.pAB * a10_frac * aSB0_conc
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - kdAA * s.pAA * aSA1_conc
        - kdBA * s.pBA * aSA0_conc
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_1
        - RtAB_1
    )

    # Active B sequences (1st order moment)
    dydt.aSB1 = s.V * (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +kpBB * s.B_conc * (s.R_conc + aSB0_conc + aSB1_conc)
        + kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - kpBB * s.B_conc * aSB1_conc
        - kpBA * s.A_conc * (s.RB_conc + aSB1_conc)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + kdBB * s.pBB * (aSB1_conc - aSB0_conc)
        + kdBA * s.pBA * b10_frac * aSA0_conc
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - kdBB * s.pBB * aSB1_conc
        - kdAB * s.pAB * aSB0_conc
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_1
        - RtBA_1
    )

    # Active A sequences (2nd order moment)
    dydt.aSA2 = s.V * (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +kpAA * s.A_conc * (s.R_conc + aSA0_conc + 2 * aSA1_conc + aSA2_conc)
        + kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - kpAA * s.A_conc * aSA2_conc
        - kpAB * s.B_conc * (s.RA_conc + aSA2_conc)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + kdAA * s.pAA * (aSA2_conc - 2 * aSA1_conc + aSA0_conc)
        + kdAB * s.pAB * a20_frac * aSB0_conc
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - kdAA * s.pAA * aSA2_conc
        - kdBA * s.pBA * aSA0_conc
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_2
        - RtAB_2
    )

    # Active B sequences (2nd order moment)
    dydt.aSB2 = s.V * (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +kpBB * s.B_conc * (s.R_conc + aSB0_conc + 2 * aSB1_conc + aSB2_conc)
        + kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - kpBB * s.B_conc * aSB2_conc
        - kpBA * s.A_conc * (s.RB_conc + aSB2_conc)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + kdBB * s.pBB * (aSB2_conc - 2 * aSB1_conc + aSB0_conc)
        + kdBA * s.pBA * b20_frac * aSA0_conc
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - kdBB * s.pBB * aSB2_conc
        - kdAB * s.pAB * aSB0_conc
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_2
        - RtBA_2
    )

    # Inactive A sequences (0th order moment)
    dydt.iSA0 = s.V * (
        # Formation by cross-propagation (PA + B -> PB)
        +kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.pAB * aSB0_conc
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_0
        + RtAB_0
    )

    # Inactive B sequences (0th order moment)
    dydt.iSB0 = s.V * (
        # Formation by cross-propagation (PB + A -> PA)
        +kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.pBA * aSA0_conc
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_0
        + RtBA_0
    )

    # Inactive A sequences (1st order moment)
    dydt.iSA1 = s.V * (
        # Formation by cross-propagation (PA + B -> PB)
        +kpAB * s.B_conc * (s.RA_conc + aSA1_conc)
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.pAB * a10_frac * aSB0_conc
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_1
        + RtAB_1
    )

    # Inactive B sequences (1st order moment)
    dydt.iSB1 = s.V * (
        # Formation by cross-propagation (PB + A -> PA)
        +kpBA * s.A_conc * (s.RB_conc + aSB1_conc)
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.pBA * b10_frac * aSA0_conc
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_1
        + RtBA_1
    )

    # Inactive A sequences (2nd order moment)
    dydt.iSA2 = s.V * (
        # Formation by cross-propagation (PA + B -> PB)
        +kpAB * s.B_conc * (s.RA_conc + aSA2_conc)
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.pAB * a20_frac * aSB0_conc
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_2
        + RtAB_2
    )

    # Inactive B sequences (2nd order moment)
    dydt.iSB2 = s.V * (
        # Formation by cross-propagation (PB + A -> PA)
        +kpBA * s.A_conc * (s.RB_conc + aSB2_conc)
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.pBA * b20_frac * aSA0_conc
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_2
        + RtBA_2
    )

    return dydt



from .model import (
    RateCoefficients,
    ModelState,
    SequenceModelState,
    Model,
    SequenceModel,
    get_sms,
)
from .equations import kinetic_odes, Context, sequence_odes

__all__ = [
    "RateCoefficients",
    "ModelState",
    "SequenceModelState",
    "kinetic_odes",
    "Context",
    "sequence_odes",
    "Model",
    "SequenceModel",
    "get_sms",
]



from ..model import ModelState, RateCoefficients


def k_CRP2_from_ratios(
    rA: float,
    rB: float,
    rX: float = 1.0,
    KAA: float = 0.0,
    KAB: float = 0.0,
    KBA: float = 0.0,
    KBB: float = 0.0,
    kpAA: float = 1.0,
) -> RateCoefficients:

    # rX = kpAA / kpBB
    kpBB = kpAA / rX

    kpAB = kpAA / rA
    kpBA = kpBB / rB

    kdAA = KAA * kpAA
    kdAB = KAB * kpAB
    kdBA = KBA * kpBA
    kdBB = KBB * kpBB

    return RateCoefficients(
        kpAA=kpAA,
        kpAB=kpAB,
        kpBA=kpBA,
        kpBB=kpBB,
        kdAA=kdAA,
        kdAB=kdAB,
        kdBA=kdBA,
        kdBB=kdBB,
    )


def k_CRP2_equal_from_ratio(
    rA: float,
    rB: float,
    K: float,
    **kwargs,
) -> RateCoefficients:

    return k_CRP2_from_ratios(rA=rA, rB=rB, KAA=K, KAB=K, KBA=K, KBB=K, **kwargs)


def c0_CRP2(
    fA0: float = 0.50, M0: float = 1.0, M0_I0_ratio: float = 1000.0
) -> ModelState:

    A0 = fA0 * M0
    B0 = (1 - fA0) * M0
    R0 = M0 / M0_I0_ratio

    return ModelState(R=R0, A=A0, B=B0)



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

from sparks.models import SimulationData
from sparks.simulators import DeterministicSimulator
from sparks.data.initiators import AIBN
from sparks.data.monomers import methyl_methacrylate
from sparks.utils.formulas import C_to_K

from ..model import Model, RateCoefficients, ModelState


def k_MMA(temp_C: float) -> RateCoefficients:

    temp_K = C_to_K(temp_C)
    iprops = AIBN(temp_K)
    mprops = methyl_methacrylate(temp_K)

    return RateCoefficients(
        T_C=temp_C,
        iprops=iprops,
        mprops=mprops,
    )


def c0_MMA(temp_C: float, c_AIBN: float) -> ModelState:

    temp_K = C_to_K(temp_C)
    mprops = methyl_methacrylate(temp_K)

    V_basis = 1.0  # Volume in L for the basis state
    I = c_AIBN * V_basis
    M = mprops.dens_m / mprops.MW * V_basis  # Convert g/L to mol

    return ModelState(V=V_basis, I=I, M=M)


def get_validation_data():

    model = Model()

    c_AIBN = 0.0258  # mol/L

    T_C = 50
    ds = DeterministicSimulator(model, k=k_MMA(T_C))
    data_50 = ds.simulate(
        t_eval=np.linspace(0, 400, 100),
        init_state=c0_MMA(T_C, c_AIBN=c_AIBN),
    )

    T_C = 70
    ds = DeterministicSimulator(model, k=k_MMA(T_C))
    data_70 = ds.simulate(
        t_eval=np.linspace(0, 100, 100),
        init_state=c0_MMA(T_C, c_AIBN=c_AIBN),
    )

    T_C = 90
    ds = DeterministicSimulator(model, k=k_MMA(T_C))
    data_90 = ds.simulate(
        t_eval=np.linspace(0, 50, 100),
        init_state=c0_MMA(T_C, c_AIBN=c_AIBN),
    )

    return data_50, data_70, data_90


def plot_conv_validation(
    data_50: SimulationData[ModelState],
    data_70: SimulationData[ModelState],
    data_90: SimulationData[ModelState],
):
    fig, ax = plt.subplots(figsize=(4, 3.5), dpi=150)
    ax.tick_params(axis="both", direction="in", top=True, right=True)

    import pandas as pd

    true_data_50C = pd.read_csv(
        "/home/devon/Documents/Github/React/Manuscript/data/MMA/MMA_50C_Conv_Model.csv"
    )
    true_data_70C = pd.read_csv(
        "/home/devon/Documents/Github/React/Manuscript/data/MMA/MMA_70C_Conv_Model.csv"
    )
    true_data_90C = pd.read_csv(
        "/home/devon/Documents/Github/React/Manuscript/data/MMA/MMA_90C_Conv_Model.csv"
    )

    plt.plot(
        data_50.trajectory.t,
        data_50.trajectory.x(data_50.init_state),
        label="50 C",
        color="tab:blue",
    )
    plt.plot(
        data_70.trajectory.t,
        data_70.trajectory.x(data_70.init_state),
        label="70 C",
        color="tab:green",
    )
    plt.plot(
        data_90.trajectory.t,
        data_90.trajectory.x(data_90.init_state),
        label="90 C",
        color="tab:orange",
    )

    plt.plot(
        true_data_50C["Time (min)"],
        true_data_50C["Conversion (%)"],
        "o",
        color="tab:blue",
        alpha=0.25,
    )

    plt.plot(
        true_data_70C["Time (min)"],
        true_data_70C["Conversion (%)"],
        "o",
        color="tab:green",
        alpha=0.25,
    )

    plt.plot(
        true_data_90C["Time (min)"],
        true_data_90C["Conversion (%)"],
        "o",
        color="tab:orange",
        alpha=0.25,
    )

    plt.xticks(
        [0, 50, 100, 150, 200, 250, 300, 350, 400],
    )
    plt.xlim(0, 400)
    plt.ylim(0, 1)
    plt.xlabel("Time (min)")
    plt.ylabel("Monomer conversion")
    plt.legend()
    plt.show()


def plot_mw_validation(
    data_50: SimulationData[ModelState],
    data_70: SimulationData[ModelState],
    data_90: SimulationData[ModelState],
):

    MW = methyl_methacrylate(1).MW

    t_50 = data_50.trajectory.t
    t_70 = data_70.trajectory.t
    t_90 = data_90.trajectory.t
    x_50 = data_50.trajectory.x(data_50.init_state)
    x_70 = data_70.trajectory.x(data_70.init_state)
    x_90 = data_90.trajectory.x(data_90.init_state)
    Mn_50 = MW * data_50.trajectory.mu_1 / data_50.trajectory.mu_0
    Mw_50 = MW * data_50.trajectory.mu_2 / data_50.trajectory.mu_1
    Mn_70 = MW * data_70.trajectory.mu_1 / data_70.trajectory.mu_0
    Mw_70 = MW * data_70.trajectory.mu_2 / data_70.trajectory.mu_1
    Mn_90 = MW * data_90.trajectory.mu_1 / data_90.trajectory.mu_0
    Mw_90 = MW * data_90.trajectory.mu_2 / data_90.trajectory.mu_1

    fig, axs = plt.subplots(1, 3, figsize=(11, 3.5), dpi=150)
    axs[0].plot(
        x_50,
        Mn_50 / 1e5,
        label="50 C",
        color="tab:blue",
    )
    axs[0].plot(
        x_50,
        Mw_50 / 1e5,
        label="50 C",
        color="tab:blue",
        linestyle="--",
    )
    axs[1].plot(
        x_70,
        Mn_70 / 1e5,
        label="70 C",
        color="tab:green",
    )
    axs[1].plot(
        x_70,
        Mw_70 / 1e5,
        label="Mw 70 C",
        color="tab:green",
        linestyle="--",
    )
    axs[2].plot(
        x_90,
        Mn_90 / 1e5,
        label="90 C",
        color="tab:orange",
    )
    axs[2].plot(
        x_90,
        Mw_90 / 1e5,
        label="Mw 90 C",
        color="tab:orange",
        linestyle="--",
    )

    true_Mn_70C = pd.read_csv(
        "/home/devon/Documents/Github/React/Manuscript/data/MMA/MMA_70C_Mn_Model.csv"
    )
    true_Mw_70C = pd.read_csv(
        "/home/devon/Documents/Github/React/Manuscript/data/MMA/MMA_70C_Mw_Model.csv"
    )
    true_Mn_90C = pd.read_csv(
        "/home/devon/Documents/Github/React/Manuscript/data/MMA/MMA_90C_Mn_Model.csv"
    )
    true_Mw_90C = pd.read_csv(
        "/home/devon/Documents/Github/React/Manuscript/data/MMA/MMA_90C_Mw_Model.csv"
    )

    axs[1].plot(
        true_Mn_70C["Conversion (%)"],
        true_Mn_70C["Mn (Da)"] / 1e5,
        "o",
        color="tab:green",
        alpha=0.25,
    )
    axs[1].plot(
        true_Mw_70C["Conversion (%)"],
        true_Mw_70C["Mw (Da)"] / 1e5,
        "o",
        color="tab:green",
        alpha=0.25,
    )
    axs[2].plot(
        true_Mn_90C["Conversion (%)"],
        true_Mn_90C["Mn (Da)"] / 1e5,
        "o",
        color="tab:orange",
        alpha=0.25,
    )
    axs[2].plot(
        true_Mw_90C["Conversion (%)"],
        true_Mw_90C["Mw (Da)"] / 1e5,
        "o",
        color="tab:orange",
        alpha=0.25,
    )

    for ax in axs:
        ax.set_xlabel("Monomer conversion")
        ax.set_ylabel("Mn or Mw (10$^5$ Da)")
        ax.tick_params(axis="both", direction="in", top=True, right=True)
    plt.tight_layout()
    plt.show()



from typing import Type
from pathlib import Path
from dataclasses import dataclass

import numpy as np
from runkmc.simulation import RunKMC, SimulationConfig, SimulationResult

from sparks.models import (
    RateType,
    StateType,
    SimulationData,
    SupportsStochasticSimulation,
)
from sparks.simulators import Simulator


@dataclass
class KMCConfig:

    num_particles: int
    termination_time: float
    analysis_time: float


class StochasticSimulator(Simulator):
    """Kinetic Monte Carlo Simulator (using RunKMC)"""

    def __init__(
        self, model: SupportsStochasticSimulation, k: RateType, data_dir: Path | str
    ):

        super().__init__(model, k)
        self.model = model

        self.model_state_class: Type[StateType] = self.model.StateClass
        self.run_kmc = RunKMC(model.name, data_dir, force_compile=False)

    def get_simulation_config(self, model_name: str, **kwargs) -> SimulationConfig:

        kmc_config = KMCConfig(
            num_particles=self.num_particles,
            termination_time=self.t_eval[-1],
            analysis_time=self.t_eval[1] - self.t_eval[0],
        )
        kmc_inputs = self.model.get_kmc_inputs(
            self.k, self.y0, kmc_config, model_name=model_name, **kwargs
        )

        return SimulationConfig(model_name, kmc_inputs, **kwargs)

    def simulate(
        self,
        t_eval: np.ndarray,
        init_state: StateType,
        num_particles: int,
        model_name: str,
        **kwargs,
    ) -> SimulationData[StateType]:

        overwrite = kwargs.pop("overwrite", False)

        self.y0 = init_state
        self.t_eval = t_eval
        self.num_particles = num_particles
        self.model_name = model_name

        config = self.get_simulation_config(self.model_name, **kwargs)

        self.result: SimulationResult = self.run_kmc.run_or_retrieve(config, overwrite)

        data, mdata = self.model.parse_kmc_outputs(self.result, self.model_name)
        self.mdata = mdata

        return data
